# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union
from typing_extensions import Annotated, TypeAlias

from ...._utils import PropertyInfo
from .file_path_annotation import FilePathAnnotation
from .file_citation_annotation import FileCitationAnnotation

__all__ = ["Annotation"]

Annotation: TypeAlias = Annotated[Union[FileCitationAnnotation, FilePathAnnotation], PropertyInfo(discriminator="type")]



import time
from damei.nn.uaii.stream.base_queue import AbstractQue


class AbstractOutput(object):
    name = 'default name'
    status = 'stopped'
    description = 'default description'

    def __init__(self, maxlen=5, *args, **kwargs):
        self.que = AbstractQue(maxlen, *args, **kwargs)
        self.timeout = 5

    def __len__(self):
        return len(self.que)

    def __iter__(self):
        return self

    def __next__(self):
        return self.get_last()

    def push(self, data, que_wait=False):
        wait_time = que_wait if isinstance(que_wait, int) else 0.0001
        if que_wait:
            while True:
                if len(self.que) <= 3:
                    self.que.append(data)
                    break
                time.sleep(wait_time)
        else:
            self.que.append(data)
            time.sleep(wait_time)

    def pop(self, wait=True, timeout=None):
        """取出"""
        timeout = timeout if timeout else self.timeout
        if self.status == 'done':
            return 'done'
        if len(self.que):
            return self.que.pop()
        else:
            if wait:
                t0 = time.time()
                while True:
                    if len(self.que):
                        return self.que.pop()

                    if time.time() - t0 > timeout:
                        raise TimeoutError(f'Time out of {timeout}s while pop que.')
                    time.sleep(0.00001)
            else:
                return None

    def get_last(self, wait=True, timeout=None):
        """全部取出，只返回最后一个"""
        timeout = timeout if timeout else self.timeout
        if self.status == 'done':
            return 'done'
        data = None
        if len(self.que):
            while len(self.que):
                data = self.que.popleft()
            return data
        else:
            if wait:
                t0 = time.time()
                while True:
                    if len(self.que):
                        assert len(self.que) == 1
                        return self.que.pop()
                    if time.time() - t0 > timeout:
                        raise TimeoutError(f'Time out of {timeout}s while pop que.')
                    time.sleep(0.00001)
            else:
                return None

    def scan(self, last=False):
        """窥视"""
        if self.status == 'done':
            return 'done'
        if len(self.que):
            tmp = self.que[-1] if last else self.que[0]
            return tmp
        else:
            return None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from .._models import BaseModel
from .file_object import FileObject

__all__ = ["Upload"]


class Upload(BaseModel):
    id: str
    """The Upload unique identifier, which can be referenced in API endpoints."""

    bytes: int
    """The intended number of bytes to be uploaded."""

    created_at: int
    """The Unix timestamp (in seconds) for when the Upload was created."""

    expires_at: int
    """The Unix timestamp (in seconds) for when the Upload was created."""

    filename: str
    """The name of the file to be uploaded."""

    object: Literal["upload"]
    """The object type, which is always "upload"."""

    purpose: str
    """The intended purpose of the file.

    [Please refer here](https://platform.openai.com/docs/api-reference/files/object#files/object-purpose)
    for acceptable values.
    """

    status: Literal["pending", "completed", "cancelled", "expired"]
    """The status of the Upload."""

    file: Optional[FileObject] = None
    """The ready File object after the Upload is completed."""



"""
测试
"""

# ! /usr/bin/env python
# coding=utf8
import grpc
# from gRPC_example import hello_pb2_grpc, hello_pb2
import hello_pb2_grpc, hello_pb2
# ! /usr/bin/env python
# coding=utf8
import time
from concurrent import futures


_ONE_DAY_IN_SECONDS = 60 * 60 * 24


class TestService(hello_pb2_grpc.GrpcServiceServicer):
    '''
    继承GrpcServiceServicer,实现hello方法
    '''

    def __init__(self):
        pass

    def hello(self, request, context):
        '''
        具体实现hello的方法，并按照pb的返回对象构造HelloResponse返回
        :param request:
        :param context:
        :return:
        '''
        result = request.data + request.skill.name + " this is gprc test service"
        list_result = {"12": 1232}
        return hello_pb2.HelloResponse(result=str(result),
                                       map_result=list_result)


def run():
    '''
    模拟服务启动
    :return:
    '''
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    hello_pb2_grpc.add_GrpcServiceServicer_to_server(TestService(), server)
    server.add_insecure_port('[::]:50052')
    server.start()
    print("start service...")
    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        server.stop(0)


if __name__ == '__main__':
    run()



from typing import List, Union, Optional
from pydantic import BaseModel, RootModel, ConfigDict

class ImageContent(BaseModel):
    type: str
    image_url: dict

class TextContent(BaseModel):
    type: str
    text: str

class MessageContent(RootModel):
    root: Union[TextContent, ImageContent]

class OpenAIChatMessage(BaseModel):
    role: str
    content: Union[str, List[MessageContent]]

    model_config = ConfigDict(extra="allow")

class OpenAIChatCompletionForm(BaseModel):
    stream: bool = True
    model: str
    messages: List[OpenAIChatMessage]

    model_config = ConfigDict(extra="allow")

class FilterForm(BaseModel):
    body: dict
    user: Optional[dict] = None
    model_config = ConfigDict(extra="allow")


import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import AsyncGenerator, Generator

def sync_wrapper(async_gen: AsyncGenerator) -> Generator:
    loop = asyncio.new_event_loop()
    with ThreadPoolExecutor(1) as executor:
        while True:
            try:
                yield executor.submit(
                    lambda: loop.run_until_complete(async_gen.__anext__())
                ).result()
            except StopAsyncIteration:
                break
    loop.close()




# import types
# from . import resources
# from ._related_class import WorkerInfo
# from typing import Union

# class LRemoteModel:
#     """
#     Local Remote Model
#     """

#     def __init__(
#             self,
#             name: str,
#             worker_info: WorkerInfo,
#             worker_resource: Union[resources.AsyncWorker, resources.Worker]
#             ) -> None:
#         self.name = name
#         self.wr: Union[resources.AsyncWorker, resources.Worker] = worker_resource
#         if not isinstance(worker_info, WorkerInfo):
#             raise ValueError(f"Failed to get remote model: {worker_info}")

#         self.model_resource = worker_info.get_model_resource(model_name=name)
#         self.model_functions = self.model_resource.model_functions
#         # self.model_functions = ["train", "__call__"]
#         if len(self.model_functions) == 0:
#             raise ValueError(f"Remote model `{self.name}` has no functions can be called remotely, please check the worker model.")

#         # 自动注册远程模型的函数
#         for func in self.model_functions:
#             original_func = self.function_warpper(self.name, func)
#             # Use functools.wraps to preserve original function metadata when creating new methods
#             # @functools.wraps(original_func)
#             # def wrapper(*args, **kwargs):
#             #     return original_func(*args, **kwargs)
#             # setattr(self, func, types.MethodType(wrapper, self))
#             setattr(self, func, types.MethodType(original_func, self))
   

#     def function_warpper(self, model_name: str, function_name: str):
#         def call_remote_function(*args, **kwargs):
#             if isinstance(args[0], LRemoteModel):
#                 # 为了处理通过types.MethodType注册时，第一个参数是self的情况
#                 args = args[1:]
#             rst =  self.wr.request(
#                 target={
#                     "model": model_name, 
#                     "function": function_name
#                     },
#                 args=args,
#                 kwargs=kwargs,
#             )
#             return rst
#         return call_remote_function
    
#     def __call__(self, *args, **kwargs):
#         return self.function_warpper(self.name, '__call__')(*args, **kwargs)

#     def functions(self):
#         return self.model_functions


# class LRModel(LRemoteModel):
#     """
#     Alias of Local Remote Model
#     """
#     ...

import os
import types
from typing import Union
from . import resources
from ._related_class import WorkerInfo

class LRemoteModel:
    """
    Local Remote Model
    """
    def __init__(
            self,
            name: str,
            worker_info: WorkerInfo,
            worker_resource: Union[resources.AsyncWorker, resources.Worker]
            ) -> None:
        
        self.name = name
        self.wr: Union[resources.AsyncWorker, resources.Worker] = worker_resource
        if not isinstance(worker_info, WorkerInfo):
            raise ValueError(f"Failed to get remote model: {worker_info}")
        self.model_resource = worker_info.get_model_resource(model_name=name)
        self.model_functions = self.model_resource.model_functions
        
        # 如果远程模型没有可调用函数，则抛出异常
        if len(self.model_functions) == 0:
            raise ValueError(f"Remote model `{self.name}` has no functions that can be called remotely. Please check the worker model.")

        # 根据资源类型（同步/异步）注册远程函数
        for func in self.model_functions:
            method = self._create_remote_method(name, func)
            setattr(self, func, types.MethodType(method, self))

        # 同样，根据资源类型为 __call__ 方法绑定同步或异步实现
        call_method = self._create_remote_method(name, '__call__')
        setattr(self, '__call__', types.MethodType(call_method, self))

    def _create_remote_method(self, model_name: str, function_name: str):
        """
        根据是同步 Worker 还是异步 AsyncWorker，创建相应的调用方法。
        """
        if isinstance(self.wr, resources.AsyncWorker):
            # 异步调用
            async def async_call_remote_function(*args, **kwargs):
                # 当通过 types.MethodType 绑定时，第一个参数是 self
                if args and isinstance(args[0], LRemoteModel):
                    args = args[1:]
                rst = await self.wr.request(
                    target={
                        "model": model_name,
                        "function": function_name
                    },
                    args=args,
                    kwargs=kwargs,
                )
                return rst
            return async_call_remote_function
        else:
            # 同步调用
            def sync_call_remote_function(*args, **kwargs):
                if args and isinstance(args[0], LRemoteModel):
                    args = args[1:]
                rst = self.wr.request(
                    target={
                        "model": model_name,
                        "function": function_name
                    },
                    args=args,
                    kwargs=kwargs,
                )
                return rst
            return sync_call_remote_function
    
    @property
    def functions(self):
        """
        获取远程模型提供的所有可调用函数名称。
        """
        return self.model_functions
    
    def connect(
        name: str,
        base_url: str = "https://aiapi.ihep.ac.cn/apiv2",
        api_key: str = os.getenv("HEPAI_API_KEY"),
        **kwargs,
    )-> "LRemoteModel":
        """
        Connect to hepai resources. such as remote model, llm, etc.
        Args:
            name (str): The name of the resource to connect to.
            base_url (str): The base URL of the resource.
        """
        from hepai import HepAI
        assert name, "Please provide the model name, you can use `client.models.list()` to get the model name."
        client = HepAI(
            base_url=base_url,
            api_key=api_key,
            **kwargs,
            )
        model: LRemoteModel = client.get_remote_model(model_name=name)
        return model


class LRModel(LRemoteModel):
    """
    Alias of Local Remote Model
    """
    ...
    
class RemoteModel(LRemoteModel):
    """
    Alias of Local Remote Model
    """
    ...


# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from ...._models import BaseModel

__all__ = ["RefusalDeltaBlock"]


class RefusalDeltaBlock(BaseModel):
    index: int
    """The index of the refusal part in the message."""

    type: Literal["refusal"]
    """Always `refusal`."""

    refusal: Optional[str] = None



from dataclasses import dataclass, field
from typing import Optional, Any, Literal
from .base_assistant import BaseObject

@dataclass
class MessageCreation:
    message_id: str

    def to_dict(self):
        return self.__dict__

@dataclass
class StepDetails:
    type: str
    message_creation: MessageCreation 

    def to_dict(self):
        return {
            "type": self.type,
            "message_creation": self.message_creation.to_dict()
        }

@dataclass
class Usage:
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int

    def to_dict(self):
        return self.__dict__

@dataclass
class ThreadRunStep(BaseObject):
    id: str
    object: str
    created_at: int
    run_id: str
    assistant_id: str
    thread_id: str
    type: Literal["message_creation", "tool_calls"]
    status: str

    cancelled_at: Optional[int] = None
    completed_at: Optional[int] = None
    expired_at: Optional[int] = None
    failed_at: Optional[int] = None
    last_error: Optional[str] = None
    step_details: StepDetails = None
    usage: Usage = None

    @property
    def output_keys(self):
        return [
            "id", "object", "created_at", "run_id", "assistant_id", "thread_id", "type", "status",
            "cancelled_at", "completed_at", "expired_at", "failed_at", "last_error", "step_details", "usage"
            ]
    
    @property
    def allowed_update_keys(self):
        return [
            "status", "cancelled_at", "completed_at", "expired_at", "failed_at", "last_error", "step_details", "usage"
            ]

    def to_dict(self, only_output_keys=True):
        new_dict = dict()
        for k, v in self.__dict__.items():
            if only_output_keys and k not in self.output_keys:
                continue
            if k == "step_details" and v:
                v = v.to_dict()
            elif k == "usage" and v:
                v = v.to_dict()
            new_dict[k] = v
        return new_dict

    def construct_status_event(self, status: str=None):
        """根据状态构建事件对象"""
        status = status or self.status
        return {
            "data": self.to_dict(),
            "event": f"thread.run.step.{status}"
            }

    



from dataclasses import dataclass, field
from typing import List, Optional, Dict, Literal, Generator, Union
import json, re, time
from openai.types.chat import ChatCompletionChunk
from .base_assistant import BaseObject


@dataclass
class Text:
    value: str
    annotations: List[dict] = field(default_factory=list)

    def to_dict(self):
        return {
            "value": self.value,
            "annotations": self.annotations
        }

@dataclass
class Content:
    type: Literal["text", "image", "video", "audio", "file"]
    text: Text = None

    def to_dict(self):
        return {
            "type": self.type,
            "text": self.text.to_dict() if self.text else None
        
        }

@dataclass
class ThreadMessage(BaseObject):
    id: str
    object: str
    created_at: int
    assistant_id: Optional[str] 
    thread_id: str
    run_id: Optional[str]
    role: str
    content: List[Content]

    order_id: Optional[str]
    attachments: List[dict] = field(default_factory=list)
    metadata: Dict[str, str] = field(default_factory=dict)

    sender: Optional[str] = None  # Indicate who created this message, e.g. specific username or assistant name
    task_type: Optional[str] = None  # Indicate the task type of this message, e.g. "draw", "translate", "summarize", etc.

    def __post_init__(self):
        super(ThreadMessage, self).__post_init__()
        self._thread = None
        pass

    @property
    def thread(self):
        return self._thread
    
    @thread.setter
    def thread(self, value):
        self._thread = value

    def to_dict(self, only_output_keys=True):
        """
        由于Content字段是元素为Content对象的列表，需要特殊处理
        """
        new_dict = dict()
        for k, v in self.__dict__.items():
            if only_output_keys and k not in self.output_keys:
                continue
            if k == "content" and v:
                v = [x.to_dict() for x in v]
            new_dict[k] = v
        return new_dict

    @property
    def output_keys(self):
        return ["id", "object", "created_at", "assistant_id", "thread_id", "run_id", 
                "role", "content", "order_id", "attachments", "metadata", "sender",
                "task_type"]

    @property
    def allowed_update_keys(self):
        return ["content", "attachments", "metadata", "sender", "task_type"]
    
    def to_oai_message(self):
        """转换为OAI消息格式"""
        if self.content is None:
            return None

        assert len(self.content) == 1, "Only one content is allowed in a message"

        # print("/n/n messages:",self.content[0])
        return {
            "content": self.content[0].text.value,
            "role": self.role,
        }
    
    def content_str(self):
        assert len(self.content) == 1, "Only one content is allowed in a message"
        return self.content[0].text.value
    
    def construct_status_event(self, status: str=None):
        status = status or self.status
        return {
            "data": self.to_dict(),
            "event": f"thread.message.{status}"
        }
    
    def convert_str_to_stream_message_delta(self, text: str, chunk_size: int=5, sleep_time: float=0.1):
        '''
        Convert a string to a stream of ChatCompletionChunk.
        chunk_size: the maximum length of each chunk.
        '''     
        Tester_OUTPUT = text.split("[**Tester OUTPUT**]") # 兼容tester的流式输出
        if len(Tester_OUTPUT) == 4: # 通过[**Tester OUTPUT**]确定是tester的输出，并进行分割
            chunks = [Tester_OUTPUT[0]+Tester_OUTPUT[1]+Tester_OUTPUT[2]+Tester_OUTPUT[3]]
        else:
            chunks = [text[i:i+chunk_size] for i in range(0, len(text), chunk_size)]
        
        # print(f"chunks: {chunks}")
        for i, chunk in enumerate(chunks):
            time.sleep(sleep_time)
            data = self.construct_delta_event(chunk, id=0)
            yield f'data: {json.dumps(data)}\n\n'
        
        return text

    def convert_Generator_to_message_delta(self, stream: Generator):
        
        """
        Support for Dicts in the stream, in addition to pure str.
        目前支持yeild的类型为str或ChatCompletionChunk。 后续添加
        """
        response = {
            'content': '', 
            'role': '', 
            'function_call': None,
            'tool_calls': [{
                'id': '',
                'function': {
                    'arguments': '',
                    'name': ''
                },
                'type': '', 
                'index': 0
            }]
        }
        full_response = ""
        for i, x in enumerate(stream):
            # assert isinstance(x, ChatCompletionChunk), "The stream should be a generator of ChatCompletionChunk"
            if isinstance(x, str):
                content = x
                event_id = 0
                if "![Image](data:image;base64," in x:
                    print("TODO: base_thread_message.py: Image will not be added to messages.")
            elif isinstance(x, ChatCompletionChunk):
                content:str = x.choices[0].delta.content
                event_id = x.id
            else:
                raise ValueError("The stream should be a generator of str or ChatCompletionChunk")

            if content:
                full_response += content
                # response["content"] += content
                # send to the stream
                data = self.construct_delta_event(content, id=event_id)
                yield f'data: {json.dumps(data)}\n\n'      

        # full_response = response["content"]
        tmp_content = Content(
            type="text",
            text=Text(
                value=full_response,
                annotations=[]
                ))
        assert self.content == [], "Content should be [] when finishing the stream"
        self.content = [tmp_content]
        return full_response
    
    def convert_oai_stream_to_message_delta(self, stream: Generator):
        
        """
        Support for Dicts in the stream, in addition to pure str.
        """
        response = {
            'content': '', 
            'role': '', 
            'function_call': None,
            'tool_calls': [{
                'id': '',
                'function': {
                    'arguments': '',
                    'name': ''
                },
                'type': '', 
                'index': 0
            }]
        }

        for i, x in enumerate(stream):
            assert isinstance(x, ChatCompletionChunk), "The stream should be a generator of ChatCompletionChunk"
            content = x.choices[0].delta.content

            if content:
                response["content"] += content
                # send to the stream
                data = self.construct_delta_event(content, id=x.id)
                yield f'data: {json.dumps(data)}\n\n'      

        full_response = response["content"]
        tmp_content = Content(
            type="text",
            text=Text(
                value=full_response,
                annotations=[]
                ))
        assert self.content == [], "Content should be [] when finishing the stream"
        self.content = [tmp_content]
        return full_response
            
    def emit_delta_event(self, stream: Generator):

        message_delta_generator = self.convert_oai_stream_to_message_delta(stream)
        self.queue.put(message_delta_generator)

    def construct_delta_event(self, content: str, id: str):
        '''
        - 构造OpenAI ChatCompletionChunk, 可以根据自己的要求修改data内容, 适配智能体的特殊输出和前端的解析
            - 为了兼容图像路径/url输出, 这里构造了"image_url"字段用于解析图像url
        '''
        

        # 提取pic_url和pdf_url-这里为了兼容Tseter的输出，将图片和pdf文件路径提取出来
        pic_urls = []
        pic_pattern = r'<pic: (.*?) >'
        matches = re.findall(pic_pattern, content)
        if matches:
            for match in matches:
                if match!="None":
                    pic_urls.append(match)
        pdf_urls = []
        pdf_pattern = r'<pdf: (.*?) >'
        matches = re.findall(pdf_pattern, content)
        if matches:
            for match in matches:
                if match!="None":    
                    pdf_urls.append(match)
        data = {
                    "id": id,
                    "object": "thread.message.delta",
                    "delta": {
                        "content": [
                        {
                            "index": 0,
                            "type": "text",
                            "text": { 
                                "value": content, 
                                "annotations": [] }
                        }
                        ]
                    }
                    }
        if len(pic_urls)>0 or len(pdf_urls)>0:
            data = {
                    "id": id,
                    "object": "thread.message.delta",
                    "delta": {
                        "content": [
                        {
                            "index": 0,
                            "type": "image_url",
                            "image_url": { 
                                "text": content,
                                "url": pic_urls,
                                "pdf_url": pdf_urls,
                                "detail": "auto" }
                        }
                        ]
                    }
                    }
                    
        return {
            "data": data,
            "event": f"thread.message.delta"
        }
    
    def construct_ask_human_input_event(self):

        data = {
            "id": "msg_123",
            "object": "thread.message.delta",
            "content": [
                {
                    "index": 0,
                    "type": "text",
                    "text": {
                        "value": "[ST]Please provide your input.[END]",
                        "annotations": []
                    }
                }
            ],
        }
        return {
            "data": data,
            "event": f"thread.message.delta"
        }


        

    
   


import re


def convert_to_raw_url(github_url):
    """
    Converts a GitHub URL to a raw URL.

    Example:
    https://github.com/user/repo/blob/branch/path/to/file.ext
    becomes
    https://raw.githubusercontent.com/user/repo/branch/path/to/file.ext

    Parameters:
    github_url (str): The GitHub URL to convert.

    Returns:
    str: The converted raw URL.
    """
    # Define the regular expression pattern
    pattern = r"https://github\.com/(.+?)/(.+?)/blob/(.+?)/(.+)"

    # Use the pattern to match and extract parts of the URL
    match = re.match(pattern, github_url)

    if match:
        user_repo = match.group(1) + "/" + match.group(2)
        branch = match.group(3)
        file_path = match.group(4)

        # Construct the raw URL
        raw_url = f"https://raw.githubusercontent.com/{user_repo}/{branch}/{file_path}"
        return raw_url

    # If the URL does not match the expected pattern, return the original URL or raise an error
    return github_url



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from ..._models import BaseModel
from .assistant_tool_choice_function import AssistantToolChoiceFunction

__all__ = ["AssistantToolChoice"]


class AssistantToolChoice(BaseModel):
    type: Literal["function", "code_interpreter", "file_search"]
    """The type of the tool. If type is `function`, the function name must be set"""

    function: Optional[AssistantToolChoiceFunction] = None



import os
try:
    import cv2
    import numpy as np
except:
    pass
from pathlib import Path

import damei as dm
from copy import deepcopy



class Dataloader(object):
    def __init__(self, source=None, imgsz=None):
        self.video_suffix = ['.mp4', '.avi', '.wmv']
        self.imgsz = imgsz

        if isinstance(source, str):
            self.dataset = self.get_dataset(source)
        elif isinstance(source, list):
            datasets = [None] * len(source)
            for i, s in enumerate(source):
                datasets[i] = self.get_dataset(s)
            self.dataset = MultiDatasets(datasets)
        else:
            raise NotImplementedError(f'source "{source}" type {type(source)} not supported')

    def get_dataset(self, source):
        from .datasets import LoadStreams, LoadImages
        imgsz = self.imgsz
        video_suffix = self.video_suffix
        webcam = source.isnumeric() or source.startswith(('rtsp://', 'rtmp://', 'http://')) or source.endswith('.txt')
        if webcam:
            import torch.backends.cudnn as cudnn
            cudnn.benchmark = True  # set True to speed up constant image size inference
            dataset = LoadStreams(source, img_size=imgsz, flip=False)
            dataset.is_camera = True
            dataset.is_video = False
        else:
            dataset = LoadImages(source, img_size=imgsz)
            dataset.is_camera = False
            dataset.is_video = True if Path(source).suffix in video_suffix else False
        return dataset

    def __len__(self):
        return len(self.dataset)

    def __iter__(self):
        return self.dataset.__iter__()

    def __next__(self):
        return self.dataset.__next__()


class MultiDatasets(object):

    def __init__(self, datasets):
        self.datasets = datasets
        # print(f'multi {self.datasets}')
        self.is_camera = False
        self.is_video = False
        self.count = -1

    def __len__(self):
        datasets = self.datasets
        return int(np.mean([len(x) for x in datasets]))

    def __iter__(self):
        return self

    def __next__(self):
        datasets = self.datasets
        num = len(self.datasets)
        self.count += 1
        imgs = [None] * num
        im0s = [None] * num
        paths = [None] * num
        vid_caps = [None] * num
        for i, dset in enumerate(datasets):
            # 不同的数据集长度不一样呀可能会报错
            path, imgs[i], im0s[i], vid_caps[i] = next(dset)
            if dset.is_camera:
                im0s = im0s[0]
                path = f'{path[0]}/{self.count:0>6}.jpg'
            elif dset.is_video:
                path = f'{path}/{self.count:0>6}.jpg'
            else:
                pass
            paths[i] = path
        # print('\nxx', path)
        imgs = np.array(imgs)
        im0s = np.array(im0s)
        # print('xxx', imgs.shape, im0s.shape)
        return paths, imgs, im0s, vid_caps


def load_single_img(file, img_sz=640, return_list=False):
    if not os.path.exists(file):
        return
    im0 = cv2.imread(file)
    img, ratio, (dw, dh), recover = dm.general.letterbox(deepcopy(im0), new_shape=img_sz)
    img = img[:, :, ::-1].transpose(2, 0, 1)
    img = np.ascontiguousarray(img)

    if return_list:
        return [file], img[np.newaxis, :], [im0]
    else:
        return file, img, im0



import os, sys
import logging
from threading import Thread
import time
from pathlib import Path
import damei as dm

# try:
    # import eventlet
    # from .socket.server import SocketApp
# except Exception as e:
    # dm.EXCEPTION(Exception, e, mute=True)

from .utils.base_uaii import BaseUAII
from .stream.streams import Streams

# try:
    # from .. import MODULES, SCRIPTS, IOS
# except:
    # from damei.nn.api import MODULES, SCRIPTS, IOS
from ..apis import MODULES, SCRIPTS, IOS

from .utils.config_loader import PyConfigLoader as Config

pydir = Path(os.path.abspath(__file__)).parent
# sys.path.append(f'{pydir.parent}')
# from modules import *  # 不能用from xsensing_ai.module import *，否则在pyconfigloader里会二次注册
logger = dm.getLogger('uaii_main')


class UAII(BaseUAII):
    """UAII: Unified AI Interface.
    uaii = dm.nn.api.UAII()
    """

    def __init__(self, cfg=None, root_path=None, **kwargs):
        super(UAII, self).__init__(**kwargs)
        self.root_path = root_path if root_path else f'{pydir.parent.parent.parent}'
        cfg = cfg if cfg else f'{pydir.parent}/configs/uaii_stream_config.py'
        self.cfg = Config(cfg_file=cfg, root_path=self.root_path)
        self._modules = MODULES  # 一个对象
        self._scripts = SCRIPTS
        self._ios = IOS
        self._configure()
        # print(self.modules)
        self._streams = self.init_streams(cfg=self.cfg)
        # print(self.streams)
        # self.configure()
        self._name2cls = self.register_name2cls()  # 记录模块名和它注册到了那个模块中

    def __call__(self, *args, **kwargs):
        pass

    def init_streams(self, cfg):
        return Streams(cfg, parent=self)

    def run_xxx_deprecated(self):
        """配置管道、初始化、运行流
        :param name:
        :param pipe: 指定管道，None是默认运行全部。
        """
        uaii_cfg = self.cfg

        streams = [x for x in uaii_cfg.enable_streams if x in uaii_cfg.streams.keys()]  # 默认全部

        # 为每个流配置管道，运行流，一个管道有多个模块，每个模块可视为一截管道
        running_modules = {}  # 已经运行的模块
        for i, stream in enumerate(streams):
            stream_cfg = uaii_cfg.streams[stream]  # list, 元素长度是模块数目，前一个模块的输入是后一个模块的输出
            module_names = [x['type'] for x in stream_cfg]

            # module = None
            last_mo = None
            for j, mname in enumerate(module_names):
                mtask = stream_cfg[j].get('task', None)
                print(f'running: {mname:<15} task: {mtask if mtask else "None":<6} last mo: {last_mo}')
                if f'{mname} {mtask}' in running_modules.keys():
                    last_mo = running_modules[f'{mname} {mtask}'].mo
                    continue
                mi = None if j == 0 else last_mo  # 模块输入
                run_in_main_thread = False if mname == 'deepsort' else False
                module = self.configure_and_run_module(
                    module=mname, mtask=mtask, mi=mi,
                    run_in_main_thread=run_in_main_thread,
                )
                running_modules[f'{mname} {mtask}'] = module
                last_mo = module.mo

        # self.fetch(running_modules['deepsort infer'].mo)

        done = True
        if done:
            print(f'done')
        return done

    def fetch(self, stream):
        """输出流"""
        fetch_times = 1
        while True:
            t0 = time.time()
            if len(stream.que):
                data = stream.get_last()
                # demo_for_dm.data = stream.scan()
                print(f'fetch times: {fetch_times} que lenth: {len(stream.que)}'
                      f' demo_for_dm.data: {data[:3]}', end='')
                fetch_times += 1
                print(f' {(time.time() - t0) * 1000:.6f} ms')
            time.sleep(0.0001)
        print('')

    def start(self, stream=None, module=None):
        """开始运行某些模块/流"""
        # print(stream, module)
        ps = self.ps()
        print(ps)
        return self.run()

    def stop(self, stream=None, pipeline=None, module=None):
        """停止指定模块或停止所有模块"""
        logger.info('stop stream pipeline module')

        print(stream, pipeline, module)
        exit()

    def build_io_instance(self, io_type, m_cfg):
        """
        构建io实例
        :param io_type: input/output
        :param m_cfg: module config
            for example:
                dict(
                    type='visible_input',
                    enable=True,
                    source='/path/to/demo_for_dm.data',
                    ...,
                )
        :return: i/o instance from self._ios
        """
        if io_type in ['input_stream', 'output_stream']:
            raise NotImplementedError(f'{io_type} not implemented')

        m_io_stream_cfg = m_cfg.get(io_type, None)
        if not m_io_stream_cfg:
            return None

        module_name = m_io_stream_cfg['type']
        if module_name is None:
            return None
        enabled = m_io_stream_cfg.get('enabled', True)
        if not enabled:
            return None
        io = self.get_module(name=module_name)  # input or output
        # print(f'{io_type} module: {module_name} m_cfg: {m_cfg}')
        io = io(mcfg=m_cfg)
        return io

    def run(self, stream, **kwargs):
        return self.run_stream(stream, run_in_main_thread=True, **kwargs)

    def start_stream(self, stream, **kwargs):
        # is_req = kwargs.get('is_req', False)
        return self.run_stream(stream, run_in_main_thread=False, **kwargs)

    def run_stream(self, stream, **kwargs):
        """
        运行流
        :stream: 可以是stream_name，也可以是一个对象
        :return:
        """
        # 读取参数
        run_in_main_thread = kwargs.pop('run_in_main_thread', True)  # 是否在主进程运行
        mtask = kwargs.pop('task', 'infer')  # 任务：infer/train/evaluate
        stream_cfg = kwargs.pop('stream_cfg', None)  # 流配置
        is_req = kwargs.pop('is_req', False)  # 是否是请求
        # mi = kwargs.get('mi', None)
        # mo = kwargs.get('mo', None)

        # 读取流
        stream = self.get_stream(name=stream) if isinstance(stream, str) else stream
        # logger.info(f'Run stream "{stream.name}", params: {kwargs}.')

        # 初始化流
        if stream.status == 'stopped' or stream_cfg is not None:
            stream.init(stream_cfg=stream_cfg)  # stream_cfg不为None时，会强制初始化

        last_generator = None
        for i, (mname, m) in enumerate(stream.models.items()):
            is_first = i == 0
            is_last = i == (len(stream.models))
            # mname = module['type']
            # m = stream.modules[mname]
            # print(f'm status: {m.status}')
            assert m.status == 'ready', f'Status of "{mname}" must be "ready" rather than "{m.status}".'

            if m.status == 'running':
                logger.warn(f'Module: {mname} already run, specify "force=True" if you want to stop and run.')
                # TODO: force run_stream
                last_generator = None
                continue

            logger.info(f'Run module: {mname} mtask: {mtask} main_thread: {run_in_main_thread}')
            # 处理模块的mi
            if not is_first:  # 非第1个模块，用上一个模块的输出来覆盖mi
                # print(f'last_generator: {last_generator}')
                m.mi = last_generator
            result = self.run_module(m=m, mtask=mtask, run_in_main_thread=run_in_main_thread, **kwargs)
            # 返回值是generator，是一个生成器
            last_generator = result

        if is_req:
            if last_generator:
                return 1, f'Successfully run stream {stream.name}.'
            else:
                return 0, f'Run stream {stream.name} failed.'
        return last_generator

    def run_module(self, m, mtask=None, mi=None, mo=None, daemon=True, run_in_main_thread=False, **kwargs):
        """
        run module in stream.
        return: iterable result object if run in main thread, True/False if run in thread
        """
        if run_in_main_thread:
            retsult = self.run_module_in_main_thread(m, mtask, **kwargs)
            return retsult

        else:
            # 不指定任务时，调用call函数
            if mtask is None:
                thread = Thread(
                    target=m.__call__,
                    daemon=daemon, )
            # 推理任务
            elif mtask == 'infer':
                thread = Thread(target=m.infer, kwargs={'mi': mi, "mo": mo}, daemon=True)
            # 训练任务
            elif mtask == 'train':  # TODO
                thread = None
                m.train()
            # 评估任务
            elif mtask == 'evaluate':
                thread = None
                m.test()
            else:
                raise KeyError(f'uaii module task error. module: {m} task: {mtask}.')

            # 启动线程
            thread.start()

        return True

    def run_module_in_main_thread(self, m, mtask, **kwargs):
        """在主进程跑任务"""
        # print(f'run: {m} {mtask} {mi} {mo}')
        # 调用call
        if mtask is None:
            return m.__call__()
        # 推理任务
        elif mtask == 'infer':
            result = m.infer(**kwargs)
        # 训练任务
        elif mtask == 'train':  # TODO
            result = m.train(**kwargs)
        # 评估任务
        elif mtask == 'evaluate':
            # result = 0
            result = m.evaluate(**kwargs)
            # m.status = 'stopped'
        else:
            raise KeyError(f'uaii module task error. module: {m} task: {mtask}.')

        m.status = 'ready'
        return result

    def evaluate(self, stream, **kwargs):
        """评估某个单模块组成的流"""
        stream_name = stream if isinstance(stream, str) else stream.name
        stream_cfg = kwargs.get('cfg', None)
        # print(stream_cfg)
        logger.info(f'Evaluate monoflow: {stream_name}.')
        # cfg = kwargs.get('cfg', None)
        # print(cfg)
        result = self.run_stream(stream=stream, task='evaluate', stream_cfg=stream_cfg, **kwargs)
        return result
        # return self.run_stream(monoflow, **kwargs)

    def train(self, stream, **kwargs):
        stream_name = stream if isinstance(stream, str) else stream.name
        stream_cfg = kwargs.get('cfg', None)
        logger.info(f'Train monoflow: {stream_name}.')
        result = self.run_stream(stream=stream, task='train', stream_cfg=stream_cfg, **kwargs)
        return result

    def infer(self, stream, **kwargs):
        stream_name = stream if isinstance(stream, str) else stream.name
        stream_cfg = kwargs.get('cfg', None)
        logger.info(f'Infer monoflow: {stream_name}.')
        result = self.run_stream(stream=stream, task='infer', stream_cfg=stream_cfg, **kwargs)
        return result


def main(root_path):
    logger = logging.getLogger('uaii')
    print(f'root_path: {os.path.abspath(root_path)}')
    cfg = f'{root_path}/config/uaii_stream_config.py'
    uaii = UAII(cfg=cfg, root_path=root_path)
    # uaii.moduels.build()

    # 查看注册了的模块
    ps_info = uaii.ps()
    print(f'ps_info: \n{ps_info}')
    exit()

    # 根据配置文件，配置模块的连接关系，并初始化
    uaii.run()  # 根据配置文件运行多个管道

    print(f'modules: {uaii.modules}')
    print(f'module scope: {uaii.modules.scope}')

    # 查看注册了的模块
    ps_info = uaii.ps()
    print(f'ps_info: \n{ps_info}')

    # print(f'ms: {type(ms)} {ms}')
    # ret = ms.inference()
    # print(f'ms ret: {ret}')

    pass


def run_app(root_path):
    # build app
    logger = logging.getLogger('uais')
    cfg = f'{root_path}/config/uaii_stream_config.py'
    uaii = UAII(cfg=cfg)
    # SocketApp.uaii = uaii  # 设置socket app的unified ai service
    # socket_app = SocketApp()
    # app.logger.info('socket app加载完成')
    # eventlet.wsgi.server(eventlet.listen(('', 5000)), socket_app.app)



from autogen_ext.models.openai import OpenAIChatCompletionClient
from autogen_ext.models.openai._openai_client import (
    convert_tools, 
    _add_usage,
    to_oai_type,
    create_kwargs
    )
from autogen_ext.models.openai.config import OpenAIClientConfiguration
from autogen_core.models import ModelFamily

from typing_extensions import Unpack
import os

import warnings
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
)

from autogen_core import (
    CancellationToken,
    FunctionCall,
    Image
)
from autogen_core.models import (
    ChatCompletionTokenLogprob,
    CreateResult,
    LLMMessage,
    ModelFamily,
    RequestUsage,
    TopLogprob,
    UserMessage
)
from autogen_core.tools import Tool, ToolSchema
from openai.types.chat import (
    ParsedChoice

)
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_chunk import Choice as ChunkChoice
from pydantic import BaseModel

from autogen_ext.models._utils.normalize_stop_reason import normalize_stop_reason
from autogen_ext.models._utils.parse_r1_content import parse_r1_content

class HepAIChatCompletionClient(OpenAIChatCompletionClient):

    def __init__(self, **kwargs: Unpack[OpenAIClientConfiguration]):

        if "api_key" not in kwargs:
            kwargs["api_key"] = os.environ.get("HEPAI_API_KEY")
        if "base_url" not in kwargs:
            kwargs["base_url"] = "https://aiapi.ihep.ac.cn/apiv2"

        if "model_info" not in kwargs:
            model_info={
                "vision": False,
                "function_calling": False,  # You must sure that the model can handle function calling
                "json_output": False,
                "structured_output": False,
                "family": ModelFamily.UNKNOWN,
            }
            kwargs["model_info"] = model_info
        allowed_models = [
        "gpt-4o",
        "o1",
        "o3",
        "gpt-4",
        "gpt-35",
        "r1",
        "v3",
        "gemini-1.5-flash",
        "gemini-1.5-pro",
        "gemini-2.0-flash",
        "claude-3-haiku",
        "claude-3-sonnet",
        "claude-3-opus",
        "claude-3.5-haiku",
        "claude-3.5-sonnet"]
        for allowed_model in allowed_models:
            model = kwargs.get("model", "")
            if allowed_model in model.lower():
                if allowed_model == "v3":
                    allowed_model = "gpt-4o"
                kwargs["model_info"]["family"] = allowed_model
                kwargs["model_info"]["function_calling"] = True
                kwargs["model_info"]["json_output"] = True,
                kwargs["model_info"]["structured_output"] = True
                break

        super().__init__(**kwargs)
    


from autogen_agentchat.teams import Swarm
from autogen_agentchat.messages import BaseAgentEvent, BaseChatMessage
from autogen_agentchat.base import ChatAgent, TerminationCondition
from autogen_core import AgentRuntime
from hepai.agents.modules.managers.base_thread import Thread
from hepai.agents.modules.managers.threads_manager import ThreadsManager
from typing import List, Dict, Any


class DrSaiSwarm(Swarm):

    def __init__(
        self,
        participants: List[ChatAgent],
        termination_condition: TerminationCondition | None = None,
        max_turns: int | None = None,
        runtime: AgentRuntime | None = None,
        custom_message_types: List[type[BaseAgentEvent | BaseChatMessage]] | None = None,
        emit_team_events: bool = False,
        thread: Thread = None,
        thread_mgr: ThreadsManager = None,
        **kwargs: Any
    ) -> None:
        super().__init__(
            participants=participants,
            termination_condition=termination_condition, 
            max_turns=max_turns,
            runtime=runtime,
            custom_message_types=custom_message_types,
            emit_team_events=emit_team_events,
        )

        self._theard: Thread = thread
        self._thread_mgr: ThreadsManager = thread_mgr



from .. import uaii


def list(**kwargs):
    """
    list all algorithms
    :return: list of algorithms
    """
    info = uaii.ps(**kwargs)
    # print(info)
    return info

def list_weights(name=None, *args, **kwargs):
    """
    list weights of algorithm by name
    :param name: algorithm name
    :return: list of weights
    """
    info = uaii.list_weights(model_name=name, *args, **kwargs)
    return info


def load(name, *args, **kwargs):
    """
    load algorithm by name  
    :param name: algorithm name 
    :return: model
    """
    model = uaii.load_model(name, *args, **kwargs)
    return model
    
def docs(name):
    """
    get algorithm docs by name
    :param name: algorithm name
    :return: docs
    """
    info = f'https://ai.ihep.ac.cn/docs/{name}'
    return info






from typing import List, Dict, Union, AsyncGenerator, Type, Generator
import copy, json, asyncio
import time
from hepai.agents.modules.managers.threads_manager import ThreadsManager
THREADS_MGR = ThreadsManager()
from hepai.agents.modules.managers.base_thread import Thread

from autogen_agentchat.agents import AssistantAgent, UserProxyAgent, BaseChatAgent
from autogen_agentchat.base import Response, TaskResult
from autogen_core import FunctionCall, CancellationToken
from autogen_agentchat.messages import (
    AgentEvent,
    ChatMessage,
    TextMessage,
    ToolCallSummaryMessage,
    ToolCallRequestEvent,
    ToolCallExecutionEvent,
    ModelClientStreamingChunkEvent,
    MultiModalMessage,
    UserInputRequestedEvent,
)
from autogen_agentchat.teams import BaseGroupChat
from autogen_agentchat.ui import Console
import time

from hepai.agents.modules.managers.threads_manager import ThreadsManager
from hepai.agents.modules.managers.base_thread_message import ThreadMessage, Content, Text

from hepai.agents.utils import Logger
logger = Logger.get_logger("dr_sai.py")

import logging
# 单个模型日志
third_party_logger1 = logging.getLogger("autogen_core")
third_party_logger1.propagate = False
third_party_logger2 = logging.getLogger("autogen_agentchat.events")
third_party_logger2.propagate = False
third_party_logger3 = logging.getLogger("httpx")
third_party_logger3.propagate = False


from hepai.agents.utils.async_process import sync_wrapper
from hepai.agents.utils.oai_stream_event import (
    chatcompletionchunk, 
    chatcompletionchunkend,
    chatcompletions,
    split_string)

class DrSai:
    """
    This is the main class of OpenDrSai, in
    """
    def __init__(self, **kwargs):
        self.username = "anonymous"
        self.threads_mgr = THREADS_MGR
        self.agent_factory: callable = kwargs.pop('agent_factory', None)
        self.history_mode = kwargs.pop('history_mode', 'backend') # backend or frontend

    #### --- 关于AutoGen --- ####
    async def start_console(
            self,
            task: str,
            agent: AssistantAgent|BaseGroupChat = None, 
            **kwargs) -> Union[None, TaskResult]:
        """
        启动aotugen原生多智能体运行方式和多智能体逻辑
        """

        # agent: AssistantAgent|BaseGroupChat = self.agent_factory() if agent is None else agent
        agent: AssistantAgent | BaseGroupChat = (
            await self.agent_factory() 
            if agent is None and asyncio.iscoroutinefunction(self.agent_factory)
            else (
                self.agent_factory() 
                if agent is None 
                else agent
            )
        )

        stream = agent._model_client_stream if not isinstance(agent, BaseGroupChat) else agent._participants[0]._model_client_stream
        if stream:
            await Console(agent.run_stream(task=task))
            return 
        else:
            result:TaskResult = await agent.run(task=task)
            print(result)
            # return result
    
    #### --- 关于OpenAI Chat/Completions --- ####
    async def a_start_chat_completions(
            self, 
            **kwargs) -> AsyncGenerator:
        """
        启动聊天任务，使用completions后端模式
        加载默认的Agents, 并启动聊天任务, 这里默认使用GroupChat
        params:
        stream: bool, 是否使用流式模式
        messages: List[Dict[str, str]], 传入的消息列表
        HEPAI_API_KEY: str, 访问hepai的api_key
        usr_info: Dict, 用户信息
        base_models: Union[str, List[str]], 智能体基座模型
        chat_mode: str, 聊天模式，默认once
        **kwargs: 其他参数
        """
        
        # 从函数工厂中获取定义的Agents
        # agent: AssistantAgent|BaseGroupChat = self.agent_factory()
        agent: AssistantAgent | BaseGroupChat = (
            await self.agent_factory() 
            if asyncio.iscoroutinefunction(self.agent_factory)
            else (self.agent_factory())
        )

        # 是否使用流式模式
        agent_stream = agent._model_client_stream if not isinstance(agent, BaseGroupChat) else agent._participants[0]._model_client_stream
        stream = kwargs.pop('stream', agent_stream)
        if isinstance(agent, BaseGroupChat) and stream:
            for participant in agent._participants:
                if not participant._model_client_stream:
                    raise ValueError("Streaming mode is not supported when participant._model_client_stream is False")
        
        # 传入的消息列表
        messages: List[Dict[str, str]] = kwargs.pop('messages', [])
        usermessage = messages[-1]["content"]

        # 保存用户的extra_requests
        extra_requests: Dict = copy.deepcopy(kwargs)

        # 大模型配置
        api_key = kwargs.pop('api_key', None)
        temperature = kwargs.pop('temperature', 0.6)
        top_p = kwargs.pop('top_p', 1)
        cache_seed = kwargs.pop('cache_seed', None)
        # 额外的请求参数
        extra_body: Union[Dict, None] = kwargs.pop('extra_body', None)
        if extra_body is not None:
            ## 用户信息 从DDF2传入的
            user_info: Dict = kwargs.pop('extra_body', {}).get("user", {})
            self.username = user_info.get('email', None) or user_info.get('name', "anonymous")
            chat_id = extra_body.get("chat_id", None) # 获取前端聊天界面的session_id
        else:
            #  {'model': 'drsai_pipeline', 'user': {'name': '888', 'id': '888', 'email': 888', 'role': 'admin'}, 'metadata': {}, 'base_models': 'openai/gpt-4o', 'apikey': 'sk-88'}
             user_info = kwargs.pop('user', {})
             self.username = user_info.get('email', None) or user_info.get('name', "anonymous")
             chat_id = kwargs.pop('chat_id', None) # 获取前端聊天端口的session_id
             history_mode = kwargs.pop('history_mode', None) or self.history_mode # backend or frontend
                
        # 使用thread加载后端的聊天记录
        # TODO: 这里需要改成异步加载
        thread: Thread = self.threads_mgr.create_threads(username=self.username, dialog_id=chat_id)
        thread.metadata["extra_requests"] = extra_requests
        agent._thread = thread
        agent._thread_mgr = self.threads_mgr
        # 如果前端没有给定dialog_id，则将当前历史消息记录加入到新的thread中/或者使用前端历史消息
        if not chat_id or history_mode == "frontend":
            thread.messages = [] # 清空历史消息
            for message in messages[:-1]:
                thread_content = [Content(type="text", text=Text(value=message["content"],annotations=[]))]
                self.threads_mgr.create_message(
                    thread=thread,
                    role = message["role"],
                    content=thread_content,
                    sender=message["role"],
                    metadata={},
                    )
                
        # 将历史消息单独保存到metadata中
        history_aoi_messages = [ {"role": x.role, "content": x.content_str(), "name": x.sender} for x in thread.messages] # 不将usermessage加入到历史消息中，在智能体会重复发送
        thread.metadata["history_aoi_messages"] = history_aoi_messages

        # 由于groupchat中不能将历史消息传入队列中，因为必须由每个Agent来处理历史消息
        if isinstance(agent, BaseGroupChat):
            for participant in agent._participants:
                participant._thread = thread
                participant._thread_mgr = self.threads_mgr

        # 启动聊天任务
        res = agent.run_stream(task=usermessage)
        tool_flag = 0
        role = ""
        async for message in res:
            
            # print(message)
            oai_chunk = copy.deepcopy(chatcompletionchunk)
            # The Unix timestamp (in seconds) of when the chat completion was created
            oai_chunk["created"] = int(time.time())
            if isinstance(message, ModelClientStreamingChunkEvent):
                if stream and isinstance(agent, BaseChatAgent):
                    content = message.content
                    oai_chunk["choices"][0]["delta"]['content'] = content
                    oai_chunk["choices"][0]["delta"]['role'] = 'assistant'
                    yield f'data: {json.dumps(oai_chunk)}\n\n'
                elif stream and isinstance(agent, BaseGroupChat):
                    role_tmp = message.source
                    if role != role_tmp:
                        role = role_tmp
                        # oai_chunk["choices"][0]["delta"]['content'] = f"\n\n**Speaker: {role}**\n\n"
                        if role:
                            oai_chunk["choices"][0]["delta"]['content'] = f"\n\n**{role}发言：**\n\n"
                        oai_chunk["choices"][0]["delta"]['role'] = 'assistant'
                        yield f'data: {json.dumps(oai_chunk)}\n\n'
                    
                    content = message.content
                    oai_chunk["choices"][0]["delta"]['content'] = content
                    oai_chunk["choices"][0]["delta"]['role'] = 'assistant'
                    yield f'data: {json.dumps(oai_chunk)}\n\n'
                    
                else:
                    if stream:
                        raise ValueError("No valid agent type for chat completions")
                    else:
                        pass
            elif isinstance(message, TaskResult):
                if stream:
                    # 最后一个chunk
                    chatcompletionchunkend["created"] = int(time.time())
                    yield f'data: {json.dumps(chatcompletionchunkend)}\n\n'
            elif isinstance(message, TextMessage):
                # 将智能体回复加入thread.messages中
                self.threads_mgr.create_message(
                    thread=thread,
                    role = "assistant",
                    content=[Content(type="text", text=Text(value=message.content,annotations=[]))],
                    sender=message.source,
                    metadata={},
                    )
                chatcompletions["choices"][0]["message"]["created"] = int(time.time())
                if (not stream) and isinstance(agent, BaseChatAgent):
                    if message.source!="user":
                        content = message.content
                        chatcompletions["choices"][0]["message"]["content"] = content
                        yield f'data: {json.dumps(chatcompletions)}\n\n'
                elif (not stream) and isinstance(agent, BaseGroupChat):
                    if message.source!="user":
                        content = message.content
                        source = message.source
                        content = f"\n\nSpeaker: {source}\n\n{content}\n\n"
                        chatcompletions["choices"][0]["message"]["content"] = content
                        yield f'data: {json.dumps(chatcompletions)}\n\n'
                else:
                    if (not stream):
                        raise ValueError("No valid agent type for chat completions")
                    else:
                        pass
            elif isinstance(message, ToolCallRequestEvent):
                tool_flag = 1
                tool_content: List[FunctionCall]=message.content
                tool_calls = []
                for tool in tool_content:
                    tool_calls.append(
                        {"id": tool.id, "type": "function","function": {"name": tool.name,"arguments": tool.arguments}}
                         )
                if stream:
                    oai_chunk["choices"][0]["delta"]['tool_calls'] = tool_calls
                    oai_chunk["choices"][0]["delta"]['role'] = 'assistant'
                    yield f'data: {json.dumps(oai_chunk)}\n\n'
                else:
                    chatcompletions["choices"][0]["message"]["tool_calls"] = tool_calls
            elif isinstance(message, ToolCallExecutionEvent):
                tool_flag = 2
            elif isinstance(message, ToolCallSummaryMessage):
                # 将智能体的ToolCallSummaryMessage回复加入thread.messages中
                self.threads_mgr.create_message(
                    thread=thread,
                    role = "assistant",
                    content=[Content(type="text", text=Text(value=message.content,annotations=[]))],
                    sender=message.source,
                    metadata={},
                    )
                if tool_flag == 2:
                    if not stream:
                        content = message.content
                        chatcompletions["choices"][0]["message"]["content"] = content
                        yield f'data: {json.dumps(chatcompletions)}\n\n'
                    else:
                        oai_chunk["choices"][0]["delta"]['content'] = message.content
                        oai_chunk["choices"][0]["delta"]['role'] = 'assistant'
                        yield f'data: {json.dumps(oai_chunk)}\n\n'
                    tool_flag = 0
            # elif isinstance(message, Response):
            #     # print("Response: " + str(message))
            # elif isinstance(message, UserInputRequestedEvent):
            #     print("UserInputRequestedEvent:" + str(message))
            # elif isinstance(message, MultiModalMessage):
            #     print("MultiModalMessage:" + str(message))
            # elif isinstance(message, ThoughtEvent):
            #     print("ThoughtEvent:" + str(message))
            else:
                # print("Unknown message:" + str(message))
                # print(f"Unknown message, type: {type(message)}")
                pass


        






from typing import Dict, List, Literal, Optional, Union, Generator, Callable
from dataclasses import dataclass, field

from fastapi import FastAPI, APIRouter, Request, HTTPException
from fastapi import UploadFile, File, Form, Query, Body
from fastapi.background import BackgroundTasks
from fastapi.responses import JSONResponse, StreamingResponse, FileResponse, HTMLResponse

import asyncio
from asyncio import Semaphore
import threading
from pydantic import BaseModel
from logging import Logger
import httpx
import inspect
# import markdown

from ._worker_class import HWorkerConfig, HRemoteModel, CommonWorker, ModelResourceInfo
from ._related_class import WorkerStoppedInfo, WorkerInfoRequest, WorkerUnifiedGateRequest


class FunctionParamsItem(BaseModel):
    args: List = []
    kwargs: Dict = {}


def get_fastapi_init_params():
    # 获取 FastAPI 的 __init__ 方法签名
    init_signature = inspect.signature(FastAPI.__init__)
    # 提取参数名（排除 'self' 和可变参数）
    params = []
    for name, param in init_signature.parameters.items():
        if name == "self":
            continue
        if param.kind in (
            inspect.Parameter.VAR_POSITIONAL,  # *args
            inspect.Parameter.VAR_KEYWORD       # **kwargs
        ):
            continue
        params.append(name)
    return params

class HWorkerAPP(FastAPI):
    """
    FastAPI app for worker
    多线程运行，支持多个模型
    """

    def __init__(
            self,
            models: HRemoteModel | List[HRemoteModel], 
            worker_config: HWorkerConfig = None,  # Alias of config
            logger: Logger = None,
            **worker_overrides,
            ):
        # 获取父类的所有可能参数并传递给父类
        # 获取 FastAPI 接受的参数名
        fastapi_params = get_fastapi_init_params()
        # 筛选出有效参数
        fastapi_kwargs = {
            k: v for k, v in worker_overrides.items()
            if k in fastapi_params
        }
        # 从 worker_overrides 中删除 FastAPI 的参数
        for k in fastapi_kwargs.keys():
            worker_overrides.pop(k)
        super().__init__(**fastapi_kwargs)
        worker_config = worker_config if worker_config is not None else HWorkerConfig()
        assert isinstance(worker_config, HWorkerConfig), f"worker_config should be an instance of HWorkerConfig"
        worker_config.update_from_dict(worker_overrides)
        # 用于控制模型访问的信号量
        self.limit_model_concurrency = worker_config.limit_model_concurrency
        self.model_semaphore: Semaphore = None
        self.global_counter = 0

        self.logger = self.get_logger(logger)
        self.worker = CommonWorker(
            app=self, models=models, worker_config=worker_config, 
            logger=self.logger)
        self._init_routers()

    def _init_routers(self):
        
        router_prefix = self.worker.config_dict.get("route_prefix", "")
        index_router = APIRouter(prefix="", tags=["base"])
        index_router.get("/")(self.index)
        self.include_router(index_router)

        router = APIRouter(prefix=router_prefix, tags=["worker"])
        router.post("/worker_unified_gate/")(self.worker_unified_gate)
        router.post("/worker_unified_gate/{function}")(self.worker_unified_gate)
        router.post("/worker_unified_gate/{model}/{function}")(self.worker_unified_gate)  # 多模型模式下，需要指定模型
        router.get("/worker_get_status")(self.worker_get_status)
        router.post("/shutdown_worker")(self.shutdown_worker)
        
        router.post("/worker/get_worker_info")(self.get_worker_info)  # 这个路由是为了与controller相同的格式，使得client也能调用
        router.post("/worker/unified_gate")(self.worker_unified_gate)  # 这个路由是为了与controller相同的格式，使得client也能调用
        router.get("/models")(self.get_models)
        
        # files router
        files_router = APIRouter(prefix=router_prefix, tags=["files"])
        

        self.include_router(router)

    @classmethod
    def get_logger(cls, logger: Logger = None):
        if logger is None:
            try:
                from ...utils._logger import Logger
                logger = Logger.get_logger("worker_app.py")
            except:
                import logging
                logger = logging.getLogger("worker_app.py")
        return logger

    @property
    def host(self):
        return self.worker.network_info.host
    
    @property
    def port(self):
        return self.worker.network_info.port

    def get_queue_length(self):
        model_semaphore: asyncio.Semaphore = self.model_semaphore
        if model_semaphore is None:
            return 0
        else:
            _value = model_semaphore._value
            _num_waiters = len(model_semaphore._waiters) if model_semaphore._waiters is not None else 0
            return self.limit_model_concurrency - _value + _num_waiters
    
    def release_model_semaphore(self):
        if self.model_semaphore is not None:
            self.model_semaphore.release()

    async def index(self):
        worker_name = self.worker.worker_id

        fastapi_docs_url = f"http://{self.host}:{self.port}/docs"
        async with httpx.AsyncClient() as client:
            # 请求 FastAPI 文档的内容（假设 `/docs` 页面返回的是 HTML 内容）
            response = await client.get(fastapi_docs_url)
            docs_content = response.text

        model_names = [m.name for m in self.worker.models]
        markdown_content = f"""
# HepAI Worker API
- Worker Name: **{worker_name}**
- Models: **{model_names}**
"""
        
        mrs = self.worker.get_model_resource_info()
        for mr in mrs:
            markdown_content += f"""
### {mr.model_name}
- Type: {mr.model_type}
- Functions: {mr.model_functions}
"""
        import markdown
        html_mk_content = markdown.markdown(markdown_content)

        
        content = f"""
            <html>
                <head>
                    <title>HepAI Worker Info</title>
                </head>
                <body>
                    <div style="padding-left: 20px;">
                        {html_mk_content}
                    </div>
                    <div>
                        {docs_content}
                    </div>
                </body>
            </html>
            """
        return HTMLResponse(content=content)
    
    # async def local_worker_unified_gate(
    #         self,
    #         worker_unified_gate_request: WorkerUnifiedGateRequest,
    #         model: str,
    #         function: str = "__call__",
    #         # user_auth: HAPIKeyAuth = api_key_auth,
    #         ) -> JSONResponse:
    #     wk_ugr = worker_unified_gate_request
    #     return self.worker.worker_unified_gate()


    async def worker_unified_gate(
            self,
            function_params: FunctionParamsItem,
            model: str = None,
            function: str = "__call__",
            # file: UploadFile = File(None),
            ):
        # global model_semaphore, global_counter
        model_semaphore = self.model_semaphore  # 这个是用于获取队列长度的
        global_counter = self.global_counter
        global_counter += 1
        if model_semaphore is None:
            model_semaphore = asyncio.Semaphore(
                self.limit_model_concurrency)
        await model_semaphore.acquire()
        try:
            rst = await self.worker.unified_gate_async(
                model=model,
                function=function, 
                args=function_params.args,
                kwargs=function_params.kwargs)
        except Exception as e:  # 如果出错，也要释放锁
            self.release_model_semaphore()
            raise e
        # 如果成功，添加背景任务并返回
        background_tasks = BackgroundTasks()  # 背景任务
        background_tasks.add_task(self.release_model_semaphore)  # 释放锁
        return rst
    
    async def worker_get_status(self):
        return self.worker.get_status_info().to_dict()
    
    async def get_worker_info(
            self, 
            worker_info_request: WorkerInfoRequest,
            # user_auth: HAPIKeyAuth = api_key_auth,
            ) -> JSONResponse:
        """
        与controller的worker_info接口一致，以便client调用
        """
        return self.worker.get_worker_info()

    async def get_models(self):
        mrs: List[ModelResourceInfo] = self.worker.get_model_resource_info()
        return {
            "object": "list",
            "data": [mr.to_dict() for mr in mrs],
        }
    
    
    async def shutdown_worker(self, background_tasks: BackgroundTasks):
        """接收来自Ctrl的关闭worker的信息"""
        background_tasks.add_task(self.worker.shutdown_worker)
        self.worker._is_deleted_in_controller = True
        wid = self.worker.worker_id
        return WorkerStoppedInfo(
            id=wid, stopped=True, 
            message=f"Worker `{wid}` shutdown",
            shutdown=True,
            )

    @classmethod
    def register_worker(
            cls,
            model: HRemoteModel = None,
            worker_config: HWorkerConfig = None,
            daemon: bool = False,
            standalone: bool = False,
            **kwargs,
            ):
        """注册HModelWorker到HaiDDF"""
        from .utils import run_standlone_worker_demo
        if standalone:  # 独立程序模式，用户测试程序
            # cls().logger.info("Running `worker` in standalone mode.")
            print("Running `worker` in standalone mode.")
            return run_standlone_worker_demo()
        
        assert model is not None, f"Model should be not None"
        import uvicorn
        app: FastAPI = HWorkerAPP(model, worker_config=worker_config, **kwargs)
        def run_uvicron():
            uvicorn.run(app, host=app.host, port=app.port)

        if daemon:  # 守护进程模式，app workre在后台运行
            t = threading.Thread(target=run_uvicron, daemon=True)
            t.start()
            return app.worker.get_worker_info()
        else:  # 正常模式，app worker在前台运行
            run_uvicron()

    def run(self):
        import uvicorn
        uvicorn.run(self, host=self.host, port=self.port)










# -*- coding: utf-8 -*-
import json
from typing import List, Dict, Union, AsyncGenerator, Tuple, Any

# Model client
from hepai.agents import HepAIChatCompletionClient 

# backend thread 
from hepai.agents import Thread, ThreadsManager

# AutoGen imports
from autogen_core import CancellationToken, FunctionCall
from autogen_core.tools import BaseTool, FunctionTool, StaticWorkbench, Workbench, ToolResult, TextResultContent
from autogen_core.models import (
    LLMMessage,
    SystemMessage,
    AssistantMessage,
    UserMessage,
    ChatCompletionClient,
    CreateResult,
    FunctionExecutionResultMessage
)
from autogen_agentchat.base import Response
from autogen_agentchat.messages import (
    BaseAgentEvent,
    BaseChatMessage,
    ToolCallSummaryMessage,
    ModelClientStreamingChunkEvent,
    TextMessage,
    UserInputRequestedEvent,
    ThoughtEvent,
    HandoffMessage,
    AgentEvent,
    ChatMessage,
    MemoryQueryEvent,
    ToolCallExecutionEvent,
    ToolCallRequestEvent,
)

async def tools_reply_function( 
    oai_messages: List[str],  # OAI messages
    agent_name: str,  # Agent name
    llm_messages: List[LLMMessage],  # AutoGen LLM messages
    model_client: HepAIChatCompletionClient,  # AutoGen LLM Model client
    tools: Union[StaticWorkbench, Workbench],  # AutoGen Workbench
    cancellation_token: CancellationToken,  # AutoGen cancellation token,
    thread: Thread,  # DrSai thread
    thread_mgr: ThreadsManager,  # DrSai thread manager
    **kwargs) -> Union[str, AsyncGenerator[str, None]]:
    """
    自定义回复函数：能够智能地调用工具，然后通过工具的执行结果进行回复。
    """

    # 使用AutoGen Workbench获取工具信息
    tools_schema = []
    if isinstance(tools, StaticWorkbench):
        # tools: List[BaseTool[Any, Any]] = tools._tools
        tools_schema: List[Dict[str, Any]] = await tools.list_tools()
    
    # 进行工具执行
    model_result = None
    async for chunk in model_client.create_stream(
            messages=llm_messages,
            cancellation_token=cancellation_token,
            tools = tools_schema,
        ):
        if isinstance(chunk, CreateResult):
            model_result = chunk
        else:
            yield chunk
        
    # 进一步解析模型返回的结果
    if isinstance(model_result.content, list):
        function_calls: List[FunctionCall] = model_result.content
        function_call_contents: str = ""
        for function_call in function_calls:
            tool_result: ToolResult = await tools.call_tool(name=function_call.name, arguments=json.loads(function_call.arguments), cancellation_token=cancellation_token)
            name = tool_result.name
            content: str = "\n".join([str(i.content) for i  in tool_result.result])

            function_call_contents += f"{name}:\n{content}\n\n"

    
        prompt = f"""以下是对应工具的执行结果：
```
{function_call_contents}
```
请根据执行结果进一步回复上面用户的问题。
"""
        llm_messages.append(UserMessage(content=prompt, source="user"))
        async for chunk in model_client.create_stream(
            messages=llm_messages,
            cancellation_token=cancellation_token,
        ):
            yield chunk
    else:
        yield model_result


async def tools_recycle_reply_function( 
    oai_messages: List[str],  # OAI messages
    agent_name: str,  # Agent name
    llm_messages: List[LLMMessage],  # AutoGen LLM messages
    model_client: HepAIChatCompletionClient,  # AutoGen LLM Model client
    tools: Union[StaticWorkbench, Workbench],  # AutoGen Workbench
    cancellation_token: CancellationToken,  # AutoGen cancellation token,
    thread: Thread,  # DrSai thread
    thread_mgr: ThreadsManager,  # DrSai thread manager
    **kwargs) -> Union[str, AsyncGenerator[str, None]]:
    """
    自定义回复函数：能够智能地循环调用工具自有的工具集进行规划完成任务。
    """

    # 使用AutoGen Workbench获取工具信息
    tools_schema = []
    if isinstance(tools, StaticWorkbench):
        # tools: List[BaseTool[Any, Any]] = tools._tools
        tools_schema: List[Dict[str, Any]] = await tools.list_tools()
        tools_names = [i["name"] for i in tools_schema]

        # 将循环调用的提示词以UserMessage的形式放入
        llm_messages.append(UserMessage(content=f"""你需要回复我上面的任务，回复要求如下：
1. 如果不需要使用工具集：{tools_names}，或者工具集中的工具无法完成任务，则直接使用你的知识回复；
2. 如果我上面的任务适合使用工具集：{tools_names}，你需要根据每个工具的功能，进行任务和对应的使用工具规划，然后按照规划依次调用对应工具执行，在任务执行结束后进行总结，最后必须回复: 'TERMINATED'
""", 
source="user"))

        # 获取最大循环次数
        max_turns = kwargs.get("max_turns", 3)
        for i in range(max_turns):
            # 进行工具执行
            model_result = None
            async for chunk in model_client.create_stream(
                    messages=llm_messages,
                    cancellation_token=cancellation_token,
                    tools = tools_schema,
                ):
                if isinstance(chunk, CreateResult):
                    model_result = chunk
                else:
                    yield chunk
                
            # 进一步解析模型返回的结果
            if isinstance(model_result.content, list):
                function_calls: List[FunctionCall] = model_result.content
                function_call_contents: str = ""
                for function_call in function_calls:
                    tool_result: ToolResult = await tools.call_tool(name=function_call.name, arguments=json.loads(function_call.arguments), cancellation_token=cancellation_token)
                    name = tool_result.name
                    content: str = "\n".join([str(i.content) for i  in tool_result.result])
                    yield f"工具{name}的执行结果如下:\n"
                    yield "<think>\n"
                    yield f"{content}\n\n"
                    yield "</think>\n"
                    function_call_contents += f"工具{name}的执行结果如下:\n{content}\n\n"

                # 将执行结果以AssistMessage的形式放入
                llm_messages.append(AssistantMessage(content=function_call_contents+"如果需要继续使用工具集，请继续回复；如果不需要使用工具集，请进行总结，并回复'TERMINATE'。\n", source=agent_name))
                                                  
            # 循环调用结束，则结束对话
            elif 'TERMINATE' in model_result.content:
                yield model_result
                return
            else:
                yield model_result
                continue
    else:
        # 无工具集，则直接使用大模型进行回复
        async for chunk in model_client.create_stream(
                messages=llm_messages,
                cancellation_token=cancellation_token,
            ):
            yield chunk




from fastapi import FastAPI, Request, Depends, status, HTTPException, UploadFile, File
from fastapi.middleware.cors import CORSMiddleware
from fastapi.concurrency import run_in_threadpool


from starlette.responses import StreamingResponse, Response
from pydantic import BaseModel, ConfigDict
from typing import List, Union, Generator, Iterator


from .utils.pipelines.auth import bearer_security, get_current_user
from .utils.pipelines.main import get_last_user_message, stream_message_template
from .utils.pipelines.misc import convert_to_raw_url

from contextlib import asynccontextmanager
from concurrent.futures import ThreadPoolExecutor
from schemas import FilterForm, OpenAIChatCompletionForm
from urllib.parse import urlparse

import shutil
import aiohttp
import os
import importlib.util
import logging
import time
import json
import uuid
import sys
import subprocess


from .config import API_KEY, PIPELINES_DIR

if not os.path.exists(PIPELINES_DIR):
    os.makedirs(PIPELINES_DIR)


PIPELINES = {}
PIPELINE_MODULES = {}
PIPELINE_NAMES = {}


def get_all_pipelines():
    pipelines = {}
    for pipeline_id in PIPELINE_MODULES.keys():
        pipeline = PIPELINE_MODULES[pipeline_id]

        if hasattr(pipeline, "type"):
            if pipeline.type == "manifold":
                manifold_pipelines = []

                # Check if pipelines is a function or a list
                if callable(pipeline.pipelines):
                    manifold_pipelines = pipeline.pipelines()
                else:
                    manifold_pipelines = pipeline.pipelines

                for p in manifold_pipelines:
                    manifold_pipeline_id = f'{pipeline_id}.{p["id"]}'

                    manifold_pipeline_name = p["name"]
                    if hasattr(pipeline, "name"):
                        manifold_pipeline_name = (
                            f"{pipeline.name}{manifold_pipeline_name}"
                        )

                    pipelines[manifold_pipeline_id] = {
                        "module": pipeline_id,
                        "type": pipeline.type if hasattr(pipeline, "type") else "pipe",
                        "id": manifold_pipeline_id,
                        "name": manifold_pipeline_name,
                        "valves": (
                            pipeline.valves if hasattr(pipeline, "valves") else None
                        ),
                    }
            if pipeline.type == "filter":
                pipelines[pipeline_id] = {
                    "module": pipeline_id,
                    "type": (pipeline.type if hasattr(pipeline, "type") else "pipe"),
                    "id": pipeline_id,
                    "name": (
                        pipeline.name if hasattr(pipeline, "name") else pipeline_id
                    ),
                    "pipelines": (
                        pipeline.valves.pipelines
                        if hasattr(pipeline, "valves")
                        and hasattr(pipeline.valves, "pipelines")
                        else []
                    ),
                    "priority": (
                        pipeline.valves.priority
                        if hasattr(pipeline, "valves")
                        and hasattr(pipeline.valves, "priority")
                        else 0
                    ),
                    "valves": pipeline.valves if hasattr(pipeline, "valves") else None,
                }
        else:
            pipelines[pipeline_id] = {
                "module": pipeline_id,
                "type": (pipeline.type if hasattr(pipeline, "type") else "pipe"),
                "id": pipeline_id,
                "name": (pipeline.name if hasattr(pipeline, "name") else pipeline_id),
                "valves": pipeline.valves if hasattr(pipeline, "valves") else None,
            }

    return pipelines

def parse_frontmatter(content):
    frontmatter = {}
    for line in content.split('\n'):
        if ':' in line:
            key, value = line.split(':', 1)
            frontmatter[key.strip().lower()] = value.strip()
    return frontmatter

def install_frontmatter_requirements(requirements):
    if requirements:
        req_list = [req.strip() for req in requirements.split(',')]
        for req in req_list:
            print(f"Installing requirement: {req}")
            subprocess.check_call([sys.executable, "-m", "pip", "install", req])
    else:
        print("No requirements found in frontmatter.")

async def load_module_from_path(module_name, module_path):

    try:
        # Read the module content
        with open(module_path, 'r', encoding='utf-8') as file:
            content = file.read()

        # Parse frontmatter
        frontmatter = {}
        if content.startswith('"""'):
            end = content.find('"""', 3)
            if end != -1:
                frontmatter_content = content[3:end]
                frontmatter = parse_frontmatter(frontmatter_content)

        # Install requirements if specified
        if 'requirements' in frontmatter:
            install_frontmatter_requirements(frontmatter['requirements'])

        # Load the module
        spec = importlib.util.spec_from_file_location(module_name, module_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        print(f"Loaded module: {module.__name__}")
        if hasattr(module, "Pipeline"):
            return module.Pipeline()
        else:
            raise Exception("No Pipeline class found")
    except Exception as e:
        print(f"Error loading module: {module_name}")

        # Move the file to the error folder
        failed_pipelines_folder = os.path.join(PIPELINES_DIR, "failed")
        if not os.path.exists(failed_pipelines_folder):
            os.makedirs(failed_pipelines_folder)

        failed_file_path = os.path.join(failed_pipelines_folder, f"{module_name}.py")
        os.rename(module_path, failed_file_path)
        print(e)
    return None


async def load_modules_from_directory(directory):
    global PIPELINE_MODULES
    global PIPELINE_NAMES

    for filename in os.listdir(directory):
        if filename.endswith(".py"):
            module_name = filename[:-3]  # Remove the .py extension
            module_path = os.path.join(directory, filename)

            # Create subfolder matching the filename without the .py extension
            subfolder_path = os.path.join(directory, module_name)
            if not os.path.exists(subfolder_path):
                os.makedirs(subfolder_path)
                logging.info(f"Created subfolder: {subfolder_path}")

            # Create a valves.json file if it doesn't exist
            valves_json_path = os.path.join(subfolder_path, "valves.json")
            if not os.path.exists(valves_json_path):
                with open(valves_json_path, "w") as f:
                    json.dump({}, f)
                logging.info(f"Created valves.json in: {subfolder_path}")

            pipeline = await load_module_from_path(module_name, module_path)
            if pipeline:
                # Overwrite pipeline.valves with values from valves.json
                if os.path.exists(valves_json_path):
                    with open(valves_json_path, "r") as f:
                        valves_json = json.load(f)
                        if hasattr(pipeline, "valves"):
                            ValvesModel = pipeline.valves.__class__
                            # Create a ValvesModel instance using default values and overwrite with valves_json
                            combined_valves = {
                                **pipeline.valves.model_dump(),
                                **valves_json,
                            }
                            valves = ValvesModel(**combined_valves)
                            pipeline.valves = valves

                            logging.info(f"Updated valves for module: {module_name}")

                pipeline_id = pipeline.id if hasattr(pipeline, "id") else module_name
                PIPELINE_MODULES[pipeline_id] = pipeline
                PIPELINE_NAMES[pipeline_id] = module_name
                logging.info(f"Loaded module: {module_name}")
            else:
                logging.warning(f"No Pipeline class found in {module_name}")

    global PIPELINES
    PIPELINES = get_all_pipelines()


async def on_startup():
    await load_modules_from_directory(PIPELINES_DIR)

    for module in PIPELINE_MODULES.values():
        if hasattr(module, "on_startup"):
            await module.on_startup()


async def on_shutdown():
    for module in PIPELINE_MODULES.values():
        if hasattr(module, "on_shutdown"):
            await module.on_shutdown()


async def reload():
    await on_shutdown()
    # Clear existing pipelines
    PIPELINES.clear()
    PIPELINE_MODULES.clear()
    PIPELINE_NAMES.clear()
    # Load pipelines afresh
    await on_startup()


@asynccontextmanager
async def lifespan(app: FastAPI):
    await on_startup()
    yield
    await on_shutdown()


app = FastAPI(docs_url="/docs", redoc_url=None, lifespan=lifespan)
app.api_key = API_KEY
app.state.PIPELINES = PIPELINES


origins = ["*"]


app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.middleware("http")
async def check_url(request: Request, call_next):
    start_time = int(time.time())
    app.state.PIPELINES = get_all_pipelines()
    response = await call_next(request)
    process_time = int(time.time()) - start_time
    response.headers["X-Process-Time"] = str(process_time)

    return response


@app.get("/v1/models")
@app.get("/models")
async def get_models():
    """
    Returns the available pipelines
    """
    app.state.PIPELINES = get_all_pipelines()
    return {
        "data": [
            {
                "id": pipeline["id"],
                "name": pipeline["name"],
                "object": "model",
                "created": int(time.time()),
                "owned_by": "openai",
                "pipeline": {
                    "type": pipeline["type"],
                    **(
                        {
                            "pipelines": (
                                pipeline["valves"].pipelines
                                if pipeline.get("valves", None)
                                else []
                            ),
                            "priority": pipeline.get("priority", 0),
                        }
                        if pipeline.get("type", "pipe") == "filter"
                        else {}
                    ),
                    "valves": pipeline["valves"] != None,
                },
            }
            for pipeline in app.state.PIPELINES.values()
        ],
        "object": "list",
        "pipelines": True,
    }


@app.get("/v1")
@app.get("/")
async def get_status():
    return {"status": True}


@app.get("/v1/pipelines")
@app.get("/pipelines")
async def list_pipelines(user: str = Depends(get_current_user)):
    if user == API_KEY:
        return {
            "data": [
                {
                    "id": pipeline_id,
                    "name": PIPELINE_NAMES[pipeline_id],
                    "type": (
                        PIPELINE_MODULES[pipeline_id].type
                        if hasattr(PIPELINE_MODULES[pipeline_id], "type")
                        else "pipe"
                    ),
                    "valves": (
                        True
                        if hasattr(PIPELINE_MODULES[pipeline_id], "valves")
                        else False
                    ),
                }
                for pipeline_id in list(PIPELINE_MODULES.keys())
            ]
        }
    else:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid API key",
        )


class AddPipelineForm(BaseModel):
    url: str


async def download_file(url: str, dest_folder: str):
    filename = os.path.basename(urlparse(url).path)
    if not filename.endswith(".py"):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="URL must point to a Python file",
        )

    file_path = os.path.join(dest_folder, filename)

    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            if response.status != 200:
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail="Failed to download file",
                )
            with open(file_path, "wb") as f:
                f.write(await response.read())

    return file_path


@app.post("/v1/pipelines/add")
@app.post("/pipelines/add")
async def add_pipeline(
    form_data: AddPipelineForm, user: str = Depends(get_current_user)
):
    if user != API_KEY:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid API key",
        )

    try:
        url = convert_to_raw_url(form_data.url)

        print(url)
        file_path = await download_file(url, dest_folder=PIPELINES_DIR)
        await reload()
        return {
            "status": True,
            "detail": f"Pipeline added successfully from {file_path}",
        }
    except HTTPException as e:
        raise e
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@app.post("/v1/pipelines/upload")
@app.post("/pipelines/upload")
async def upload_pipeline(
    file: UploadFile = File(...), user: str = Depends(get_current_user)
):
    if user != API_KEY:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid API key",
        )

    file_ext = os.path.splitext(file.filename)[1]
    if file_ext != ".py":
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Only Python files are allowed.",
        )

    try:
        # Ensure the destination folder exists
        os.makedirs(PIPELINES_DIR, exist_ok=True)

        # Define the file path
        file_path = os.path.join(PIPELINES_DIR, file.filename)

        # Save the uploaded file to the specified directory
        with open(file_path, "wb") as buffer:
            shutil.copyfileobj(file.file, buffer)

        # Perform any necessary reload or processing
        await reload()

        return {
            "status": True,
            "detail": f"Pipeline uploaded successfully to {file_path}",
        }
    except HTTPException as e:
        raise e
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


class DeletePipelineForm(BaseModel):
    id: str


@app.delete("/v1/pipelines/delete")
@app.delete("/pipelines/delete")
async def delete_pipeline(
    form_data: DeletePipelineForm, user: str = Depends(get_current_user)
):
    if user != API_KEY:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid API key",
        )

    pipeline_id = form_data.id
    pipeline_name = PIPELINE_NAMES.get(pipeline_id.split(".")[0], None)

    if PIPELINE_MODULES[pipeline_id]:
        if hasattr(PIPELINE_MODULES[pipeline_id], "on_shutdown"):
            await PIPELINE_MODULES[pipeline_id].on_shutdown()

    pipeline_path = os.path.join(PIPELINES_DIR, f"{pipeline_name}.py")
    if os.path.exists(pipeline_path):
        os.remove(pipeline_path)
        await reload()
        return {
            "status": True,
            "detail": f"Pipeline {pipeline_id} deleted successfully",
        }
    else:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Pipeline {pipeline_id} not found",
        )


@app.post("/v1/pipelines/reload")
@app.post("/pipelines/reload")
async def reload_pipelines(user: str = Depends(get_current_user)):
    if user == API_KEY:
        await reload()
        return {"message": "Pipelines reloaded successfully."}
    else:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid API key",
        )


@app.get("/v1/{pipeline_id}/valves")
@app.get("/{pipeline_id}/valves")
async def get_valves(pipeline_id: str):
    if pipeline_id not in PIPELINE_MODULES:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Pipeline {pipeline_id} not found",
        )

    pipeline = PIPELINE_MODULES[pipeline_id]

    if hasattr(pipeline, "valves") is False:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Valves for {pipeline_id} not found",
        )

    return pipeline.valves


@app.get("/v1/{pipeline_id}/valves/spec")
@app.get("/{pipeline_id}/valves/spec")
async def get_valves_spec(pipeline_id: str):
    if pipeline_id not in PIPELINE_MODULES:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Pipeline {pipeline_id} not found",
        )

    pipeline = PIPELINE_MODULES[pipeline_id]

    if hasattr(pipeline, "valves") is False:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Valves for {pipeline_id} not found",
        )

    return pipeline.valves.schema()


@app.post("/v1/{pipeline_id}/valves/update")
@app.post("/{pipeline_id}/valves/update")
async def update_valves(pipeline_id: str, form_data: dict):

    if pipeline_id not in PIPELINE_MODULES:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Pipeline {pipeline_id} not found",
        )

    pipeline = PIPELINE_MODULES[pipeline_id]

    if hasattr(pipeline, "valves") is False:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Valves for {pipeline_id} not found",
        )

    try:
        ValvesModel = pipeline.valves.__class__
        valves = ValvesModel(**form_data)
        pipeline.valves = valves

        # Determine the directory path for the valves.json file
        subfolder_path = os.path.join(PIPELINES_DIR, PIPELINE_NAMES[pipeline_id])
        valves_json_path = os.path.join(subfolder_path, "valves.json")

        # Save the updated valves data back to the valves.json file
        with open(valves_json_path, "w") as f:
            json.dump(valves.model_dump(), f)

        if hasattr(pipeline, "on_valves_updated"):
            await pipeline.on_valves_updated()
    except Exception as e:
        print(e)
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"{str(e)}",
        )

    return pipeline.valves


@app.post("/v1/{pipeline_id}/filter/inlet")
@app.post("/{pipeline_id}/filter/inlet")
async def filter_inlet(pipeline_id: str, form_data: FilterForm):
    if pipeline_id not in app.state.PIPELINES:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Filter {pipeline_id} not found",
        )

    try:
        pipeline = app.state.PIPELINES[form_data.body["model"]]
        if pipeline["type"] == "manifold":
            pipeline_id = pipeline_id.split(".")[0]
    except:
        pass

    pipeline = PIPELINE_MODULES[pipeline_id]

    try:
        if hasattr(pipeline, "inlet"):
            body = await pipeline.inlet(form_data.body, form_data.user)
            return body
        else:
            return form_data.body
    except Exception as e:
        print(e)
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"{str(e)}",
        )


@app.post("/v1/{pipeline_id}/filter/outlet")
@app.post("/{pipeline_id}/filter/outlet")
async def filter_outlet(pipeline_id: str, form_data: FilterForm):
    if pipeline_id not in app.state.PIPELINES:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Filter {pipeline_id} not found",
        )

    try:
        pipeline = app.state.PIPELINES[form_data.body["model"]]
        if pipeline["type"] == "manifold":
            pipeline_id = pipeline_id.split(".")[0]
    except:
        pass

    pipeline = PIPELINE_MODULES[pipeline_id]

    try:
        if hasattr(pipeline, "outlet"):
            body = await pipeline.outlet(form_data.body, form_data.user)
            return body
        else:
            return form_data.body
    except Exception as e:
        print(e)
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"{str(e)}",
        )


@app.post("/v1/chat/completions")
@app.post("/chat/completions")
async def generate_openai_chat_completion(request: Request, form_data: OpenAIChatCompletionForm):
    headers = dict(request.headers)
    messages = [message.model_dump() for message in form_data.messages]
    user_message = get_last_user_message(messages)

    if (
        form_data.model not in app.state.PIPELINES
        or app.state.PIPELINES[form_data.model]["type"] == "filter"
    ):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Pipeline {form_data.model} not found",
        )

    def job():
        print(form_data.model)

        pipeline = app.state.PIPELINES[form_data.model]
        pipeline_id = form_data.model

        print(pipeline_id)

        if pipeline["type"] == "manifold":
            manifold_id, pipeline_id = pipeline_id.split(".", 1)
            pipe = PIPELINE_MODULES[manifold_id].pipe
        else:
            pipe = PIPELINE_MODULES[pipeline_id].pipe

        if form_data.stream:

            def stream_content():
                res = pipe(
                    user_message=user_message,
                    model_id=pipeline_id,
                    messages=messages,
                    body=form_data.model_dump(),
                    headers=headers,
                )

                # logging.info(f"stream:true:{res}")

                if isinstance(res, str):
                    message = stream_message_template(form_data.model, res)
                    # logging.info(f"stream_content:str:{message}")
                    yield f"data: {json.dumps(message)}\n\n"

                if isinstance(res, Iterator):
                    for line in res:
                        if isinstance(line, BaseModel):
                            line = line.model_dump_json()
                            line = f"data: {line}"

                        try:
                            line = line.decode("utf-8")
                        except:
                            pass

                        # logging.info(f"stream_content:Generator:{line}")

                        if line.startswith("data:"):
                            yield f"{line}\n\n"
                        else:
                            line = stream_message_template(form_data.model, line)
                            yield f"data: {json.dumps(line)}\n\n"

                if isinstance(res, str) or isinstance(res, Generator):
                    finish_message = {
                        "id": f"{form_data.model}-{str(uuid.uuid4())}",
                        "object": "chat.completion.chunk",
                        "created": int(time.time()),
                        "model": form_data.model,
                        "choices": [
                            {
                                "index": 0,
                                "delta": {},
                                "logprobs": None,
                                "finish_reason": "stop",
                            }
                        ],
                    }

                    yield f"data: {json.dumps(finish_message)}\n\n"
                    yield f"data: [DONE]"

            return StreamingResponse(stream_content(), media_type="text/event-stream")
        else:
            res = pipe(
                user_message=user_message,
                model_id=pipeline_id,
                messages=messages,
                body=form_data.model_dump(),
            )
            logging.info(f"stream:false:{res}")

            if isinstance(res, dict):
                return res
            elif isinstance(res, BaseModel):
                return res.model_dump()
            else:

                message = ""

                if isinstance(res, str):
                    message = res

                if isinstance(res, Generator):
                    for stream in res:
                        message = f"{message}{stream}"

                logging.info(f"stream:false:{message}")
                return {
                    "id": f"{form_data.model}-{str(uuid.uuid4())}",
                    "object": "chat.completion",
                    "created": int(time.time()),
                    "model": form_data.model,
                    "choices": [
                        {
                            "index": 0,
                            "message": {
                                "role": "assistant",
                                "content": message,
                            },
                            "logprobs": None,
                            "finish_reason": "stop",
                        }
                    ],
                }

    return await run_in_threadpool(job)


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description='Run the FastAPI app with a customizable port.')
    parser.add_argument('--port', type=int, default=9097, help='Port number to listen on (default: 9097)')
    args = parser.parse_args()

    import uvicorn
    # uvicorn main:app --port $PORT --host 0.0.0.0 --forwarded-allow-ips '*' --reload
    uvicorn.run("main:app", host="0.0.0.0", port=args.port, reload=True)


# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from typing_extensions import TypeAlias

from .moderation_text_input_param import ModerationTextInputParam
from .moderation_image_url_input_param import ModerationImageURLInputParam

__all__ = ["ModerationMultiModalInputParam"]

ModerationMultiModalInputParam: TypeAlias = Union[ModerationImageURLInputParam, ModerationTextInputParam]



from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials
from fastapi import HTTPException, status, Depends

from pydantic import BaseModel
from typing import Union, Optional


from passlib.context import CryptContext
from datetime import datetime, timedelta
import jwt
import logging
import os

import requests
import uuid

SESSION_SECRET = os.getenv("SESSION_SECRET", " ")
ALGORITHM = "HS256"

##############
# Auth Utils
##############

bearer_security = HTTPBearer()
pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


def verify_password(plain_password, hashed_password):
    return (
        pwd_context.verify(plain_password, hashed_password) if hashed_password else None
    )


def get_password_hash(password):
    return pwd_context.hash(password)


def create_token(data: dict, expires_delta: Union[timedelta, None] = None) -> str:
    payload = data.copy()

    if expires_delta:
        expire = datetime.utcnow() + expires_delta
        payload.update({"exp": expire})

    encoded_jwt = jwt.encode(payload, SESSION_SECRET, algorithm=ALGORITHM)
    return encoded_jwt


def decode_token(token: str) -> Optional[dict]:
    try:
        decoded = jwt.decode(token, SESSION_SECRET, algorithms=[ALGORITHM])
        return decoded
    except Exception as e:
        return None


def extract_token_from_auth_header(auth_header: str):
    return auth_header[len("Bearer ") :]


def get_current_user(
    credentials: HTTPAuthorizationCredentials = Depends(bearer_security),
) -> Optional[dict]:
    token = credentials.credentials
    return token




"""
Dr.Sai与OpenWebUI的Pipeline的API接口
"""
import os, sys
from pathlib import Path
here = Path(__file__).parent


pipe_dir = f'{here}/pipelines'


sys.path.insert(0, pipe_dir)
from .pipelines.main import app
from .pipelines.main import lifespan
sys.path.pop(0)


# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from .text import Text
from ...._models import BaseModel

__all__ = ["TextContentBlock"]


class TextContentBlock(BaseModel):
    text: Text

    type: Literal["text"]
    """Always `text`."""



from typing import Any, Callable, List, Mapping
import asyncio

from autogen_core import Component, ComponentModel, AgentRuntime
from pydantic import BaseModel
from typing_extensions import Self

from autogen_agentchat.base import ChatAgent, TerminationCondition
# from autogen_agentchat.teams import RoundRobinGroupChat
from autogen_agentchat.teams._group_chat._round_robin_group_chat import RoundRobinGroupChatManager, RoundRobinGroupChatConfig
from autogen_agentchat.messages import AgentEvent, ChatMessage, BaseAgentEvent, BaseChatMessage, MessageFactory
from autogen_agentchat.state import RoundRobinManagerState
# from autogen_agentchat.teams import BaseGroupChat
# from autogen_agentchat.teams._group_chat._base_group_chat_manager import BaseGroupChatManager
from hepai.agents.modules.groupchat._base_group_chat import DrSaiGroupChat
from hepai.agents.modules.groupchat._base_group_chat import DrSaiGroupChatManager
from autogen_agentchat.teams._group_chat._events import GroupChatTermination

from hepai.agents.modules.managers.base_thread import Thread
from hepai.agents.modules.managers.threads_manager import ThreadsManager

class DrSaiRoundRobinGroupChatManager(DrSaiGroupChatManager):
    """A group chat manager that selects the next speaker in a round-robin fashion."""

    def __init__(
        self,
        name: str,
        group_topic_type: str,
        output_topic_type: str,
        participant_topic_types: List[str],
        participant_names: List[str],
        participant_descriptions: List[str],
        output_message_queue: asyncio.Queue[BaseAgentEvent | BaseChatMessage | GroupChatTermination],
        termination_condition: TerminationCondition | None,
        max_turns: int | None,
        message_factory: MessageFactory,
        emit_team_events: bool,
        thread: Thread = None,
        thread_mgr: ThreadsManager = None,
        **kwargs: Any
    ) -> None:
       
        super().__init__(
            name=name,
            group_topic_type=group_topic_type,
            output_topic_type=output_topic_type,
            participant_topic_types=participant_topic_types,
            participant_names=participant_names,
            participant_descriptions=participant_descriptions,
            output_message_queue=output_message_queue,
            termination_condition=termination_condition,
            max_turns=max_turns,
            message_factory=message_factory,
            emit_team_events=emit_team_events,
            thread=thread,
            thread_mgr=thread_mgr,
            **kwargs
        )
        self._next_speaker_index = 0

    async def validate_group_state(self, messages: List[ChatMessage] | None) -> None:
        pass

    async def reset(self) -> None:
        self._current_turn = 0
        self._message_thread.clear()
        if self._termination_condition is not None:
            await self._termination_condition.reset()
        self._next_speaker_index = 0

    async def save_state(self) -> Mapping[str, Any]:
        state = RoundRobinManagerState(
            message_thread=list(self._message_thread),
            current_turn=self._current_turn,
            next_speaker_index=self._next_speaker_index,
        )
        return state.model_dump()

    async def load_state(self, state: Mapping[str, Any]) -> None:
        round_robin_state = RoundRobinManagerState.model_validate(state)
        self._message_thread = list(round_robin_state.message_thread)
        self._current_turn = round_robin_state.current_turn
        self._next_speaker_index = round_robin_state.next_speaker_index

    async def select_speaker(self, thread: List[AgentEvent | ChatMessage]) -> str:
        """Select a speaker from the participants in a round-robin fashion."""
        current_speaker_index = self._next_speaker_index
        self._next_speaker_index = (current_speaker_index + 1) % len(self._participant_names)
        current_speaker = self._participant_names[current_speaker_index]
        return current_speaker


class DrSaiRoundRobinGroupChat(DrSaiGroupChat):
    """A team that runs a group chat with participants taking turns in a round-robin fashion
    to publish a message to all.

    If a single participant is in the team, the participant will be the only speaker.

    Args:
        participants (List[BaseChatAgent]): The participants in the group chat.
        termination_condition (TerminationCondition, optional): The termination condition for the group chat. Defaults to None.
            Without a termination condition, the group chat will run indefinitely.
        max_turns (int, optional): The maximum number of turns in the group chat before stopping. Defaults to None, meaning no limit.

    Raises:
        ValueError: If no participants are provided or if participant names are not unique.

    Examples:

    A team with one participant with tools:

        .. code-block:: python

            import asyncio
            from autogen_ext.models.openai import OpenAIChatCompletionClient
            from autogen_agentchat.agents import AssistantAgent
            from autogen_agentchat.teams import RoundRobinGroupChat
            from autogen_agentchat.conditions import TextMentionTermination
            from autogen_agentchat.ui import Console


            async def main() -> None:
                model_client = OpenAIChatCompletionClient(model="gpt-4o")

                async def get_weather(location: str) -> str:
                    return f"The weather in {location} is sunny."

                assistant = AssistantAgent(
                    "Assistant",
                    model_client=model_client,
                    tools=[get_weather],
                )
                termination = TextMentionTermination("TERMINATE")
                team = RoundRobinGroupChat([assistant], termination_condition=termination)
                await Console(team.run_stream(task="What's the weather in New York?"))


            asyncio.run(main())

    A team with multiple participants:

        .. code-block:: python

            import asyncio
            from autogen_ext.models.openai import OpenAIChatCompletionClient
            from autogen_agentchat.agents import AssistantAgent
            from autogen_agentchat.teams import RoundRobinGroupChat
            from autogen_agentchat.conditions import TextMentionTermination
            from autogen_agentchat.ui import Console


            async def main() -> None:
                model_client = OpenAIChatCompletionClient(model="gpt-4o")

                agent1 = AssistantAgent("Assistant1", model_client=model_client)
                agent2 = AssistantAgent("Assistant2", model_client=model_client)
                termination = TextMentionTermination("TERMINATE")
                team = RoundRobinGroupChat([agent1, agent2], termination_condition=termination)
                await Console(team.run_stream(task="Tell me some jokes."))


            asyncio.run(main())
    """

    component_config_schema = RoundRobinGroupChatConfig
    component_provider_override = "autogen_agentchat.teams.RoundRobinGroupChat"

    def __init__(
        self,
        participants: List[ChatAgent],
        termination_condition: TerminationCondition | None = None,
        max_turns: int | None = None,
        runtime: AgentRuntime | None = None,
        custom_message_types: List[type[BaseAgentEvent | BaseChatMessage]] | None = None,
        emit_team_events: bool = False,
        thread: Thread = None,
        thread_mgr: ThreadsManager = None,
        **kwargs: Any
    ) -> None:
        super().__init__(
            participants=participants,
            group_chat_manager_name="DrSaiRoundRobinGroupChatManager",
            group_chat_manager_class=DrSaiRoundRobinGroupChatManager,
            termination_condition=termination_condition,
            max_turns=max_turns,
            runtime=runtime,
            custom_message_types=custom_message_types,
            emit_team_events=emit_team_events,
            thread=thread,
            thread_mgr=thread_mgr,
            **kwargs
        )

    def _create_group_chat_manager_factory(
        self,
        name: str,
        group_topic_type: str,
        output_topic_type: str,
        participant_topic_types: List[str],
        participant_names: List[str],
        participant_descriptions: List[str],
        output_message_queue: asyncio.Queue[BaseAgentEvent | BaseChatMessage | GroupChatTermination],
        termination_condition: TerminationCondition | None,
        max_turns: int | None,
        message_factory: MessageFactory,
    ) -> Callable[[], DrSaiRoundRobinGroupChatManager]:

        def _factory() -> DrSaiRoundRobinGroupChatManager:
            return DrSaiRoundRobinGroupChatManager(
                name,
                group_topic_type,
                output_topic_type,
                participant_topic_types,
                participant_names,
                participant_descriptions,
                output_message_queue,
                termination_condition,
                max_turns,
                message_factory,
                self._emit_team_events,
                thread=self._thread,
                thread_mgr=self._thread_mgr,
            )

        return _factory

    def _to_config(self) -> RoundRobinGroupChatConfig:
        participants = [participant.dump_component() for participant in self._participants]
        termination_condition = self._termination_condition.dump_component() if self._termination_condition else None
        return RoundRobinGroupChatConfig(
            participants=participants,
            termination_condition=termination_condition,
            max_turns=self._max_turns,
            emit_team_events=self._emit_team_events,
        )

    @classmethod
    def _from_config(cls, config: RoundRobinGroupChatConfig) -> Self:
        participants = [ChatAgent.load_component(participant) for participant in config.participants]
        termination_condition = (
            TerminationCondition.load_component(config.termination_condition) if config.termination_condition else None
        )
        return cls(
            participants,
            termination_condition=termination_condition,
            max_turns=config.max_turns,
            emit_team_events=config.emit_team_events,
        )







# try:
    
# except:
#     raise ImportError("openai not found, please install openai first, `pip intall openai>=1.55.0`")


# from openai._base_client import SyncAPIClient, AsyncAPIClient
# from openai._constants import DEFAULT_MAX_RETRIES, DEFAULT_TIMEOUT
# from openai._types import NOT_GIVEN, Omit
# from openai._qs import Querystring
# from openai._streaming import Stream
# from openai._exceptions import APIStatusError, OpenAIError
# from .adapted_openai.version import __version__ as __openai_version__
# from .adapted_openai._base_client import SyncAPIClient, AsyncAPIClient
# from .adapted_openai._constants import DEFAULT_MAX_RETRIES, DEFAULT_TIMEOUT
# from .adapted_openai._types import NOT_GIVEN, Omit
# from .adapted_openai._qs import Querystring
# from .adapted_openai._streaming import Stream, AsyncStream
# from .adapted_openai._exceptions import APIStatusError, OpenAIError
# from .adapted_openai import resources
# from .adapted_openai.types.chat import *



# __all__ = ["SyncAPIClient", "AsyncAPIClient", "DEFAULT_MAX_RETRIES", "DEFAULT_TIMEOUT", 
#            "NOT_GIVEN", "Omit", "Querystring", "Stream", "AsyncStream", "APIStatusError", "OpenAIError",
#            "resources", "ChatCompletion", "ChatCompletionChunk",
#            ]




from openai.version import __version__ as __openai_version__

# from openai._base_client import SyncAPIClient, AsyncAPIClient
from openai._constants import DEFAULT_MAX_RETRIES, DEFAULT_TIMEOUT
from openai._types import NOT_GIVEN, Omit
from openai._qs import Querystring
from openai._streaming import Stream, AsyncStream
from openai._exceptions import APIStatusError, OpenAIError
from openai import resources
from openai.types.chat import *

from .bridge import HAsyncAPIClient as AsyncAPIClient
from .bridge import HSyncAPIClient as SyncAPIClient







import os
from pathlib import Path
import logging
from dataclasses import dataclass

import sys
try:
    from hepai.agents.version import __version__
except:
    here = Path(__file__).parent
    sys.path.insert(1, f'{here.parent.parent}')
    from hepai.agents.version import __version__

from hepai.agents.configs import CONST

@dataclass
class LoggingLevel:
    NOTSET: int = logging.NOTSET  # 0
    DEBUG: int = logging.DEBUG  # 10
    INFO: int = logging.INFO  # 20
    WARNING: int = logging.WARNING  # 30
    WARN: int = logging.WARN  # 30
    ERROR: int = logging.ERROR  # 40
    CRITICAL: int = logging.CRITICAL  # 50
    FATAL: int = logging.FATAL  # 50
    

class DrSaiLogger:

    @classmethod
    def get_logger(
        cls, 
        name: str, 
        dir: str = CONST.LOGGER_DIR,
        level: int = CONST.LOGGER_LEVEL,
        name_length: int = 12,
        **kwargs):

        if isinstance(level, str):
            level = getattr(LoggingLevel, level.upper())

        name = f'{name:^{name_length}}'
        format_str = f"\033[1;35m[%(asctime)s]\033[0m \033[1;32m[%(name)s]\033[0m " \
                    f"\033[1;36m[%(levelname)s]:\033[0m %(message)s"
        # datafmt = '%Y-%m%d %H:%M:%S'
        logger_file = f'{dir}/{Path(os.getcwd()).name}.log'


        # logging.basicConfig(level=level,
        #                     format=format_str,
        #                     datefmt='%Y%m%d %H:%M:%S',
        #                     )
        logger = logging.getLogger(name)
        logger.setLevel(level)
        
        if not os.path.exists(dir):
            os.makedirs(dir)

        fh = logging.FileHandler(logger_file)
        fh.setLevel(level=level)
        fh.setFormatter(logging.Formatter(
            format_str, 
            # datefmt=datafmt
            ))
        
        logger.addHandler(fh)
        return logger


if __name__ == "__main__":
    logger = DrSaiLogger.get_logger(
        "DrSai",
        level="DEBUG",
        # level="INFO",
        )
    logger.info("Hello, DrSai!")
    logger.debug("Hello, DrSai!")




# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["ResponseFormatText"]


class ResponseFormatText(BaseModel):
    type: Literal["text"]
    """The type of response format being defined: `text`"""



# -*- coding=utf-8 -*-
from abc import abstractmethod


class Event(object):
    ADDED = "ADDED"
    MODIFIED = "MODIFIED"
    DELETED = "DELETED"
    pass


class AbstractListener(object):
    def __init__(self, listener_name):
        self._listener_name = listener_name

    @property
    def listener_name(self):
        return self._listener_name

    @abstractmethod
    def launch(self, *args, **kwargs):
        pass


class AbstractListenerManager(object):
    @abstractmethod
    def manager_context(self):
        pass

    @abstractmethod
    def add_listener(self, listener):
        pass

    @abstractmethod
    def remove_listener(self, listener_name):
        pass

    @abstractmethod
    def empty_listeners(self):
        pass

    @abstractmethod
    def do_launch(self):
        pass


class SubscribeListener(AbstractListener):
    def __init__(self, fn, listener_name):
        """

        :rtype: object
        """
        super(SubscribeListener, self).__init__(listener_name)
        self._fn = fn

    def launch(self, event, *args, **kwargs):
        self._fn(event, *args, **kwargs)


class SimpleListenerManager(AbstractListenerManager):

    def __init__(self):
        # listener_name --> listener
        self._listener_container = dict()

    @property
    def manager_context(self):
        return self._listener_container

    def merge_listeners(self, other_manager):
        if not other_manager or not isinstance(other_manager, AbstractListenerManager):
            return
        for listener_name, listener in other_manager.manager_context():
            self._listener_container[listener_name] = listener

    def all_listeners(self):
        return self._listener_container

    def add_listener(self, listener):
        self._listener_container[listener.listener_name] = listener
        return self

    def add_listeners(self, *listeners):
        [self.add_listener(listener) for listener in listeners]
        return self

    def remove_listener(self, listener_name):
        if listener_name not in self._listener_container.keys():
            return self
        self._listener_container.pop(listener_name)
        return self

    def empty_listeners(self):
        self._listener_container.clear()

    def do_launch(self, *args, **kwargs):
        for _, listener in self._listener_container.items():
            listener.launch(*args, **kwargs)



from .app_worker import DrSaiAPP

from autogen_agentchat.agents import AssistantAgent, UserProxyAgent, BaseChatAgent
from autogen_agentchat.teams import BaseGroupChat

from hepai import HRModel, HModelConfig, HWorkerConfig, HWorkerAPP
import hepai

import json
import os, subprocess
from dataclasses import dataclass, field
from typing import Dict, List, Union
from fastapi import FastAPI
import uvicorn
from uvicorn import Server, Config
import asyncio

class DrSaiWorkerModel(HRModel):  # Define a custom worker model inheriting from HRModel.
    def __init__(
            self, 
            config: HModelConfig,
            drsaiapp: DrSaiAPP = None # 传入DrSaiAPP实例
            ):
        super().__init__(config=config)

        # if drsaiapp is not None and isinstance(drsaiapp, type):
        #     self.drsai = drsaiapp()  # Instantiate the DrSaiAPP instance.
        # else:
        #     self.drsai = drsaiapp or DrSaiAPP()  # Instantiate the DrSaiAPP instance.
        # pass
        self.drsai = drsaiapp

    @HRModel.remote_callable  # Decorate the function to enable remote call.
    def custom_method(self, a: int = 1, b: int = 2) -> int:
        """Define your custom method here."""
        return a + b

    @HRModel.remote_callable
    def get_stream(self):
        for x in range(10):
            yield f"data: {json.dumps(x)}\n\n"

    @HRModel.remote_callable
    async def a_chat_completions(self, *args, **kwargs):
        return await self.drsai.a_start_chat_completions(*args, **kwargs)
    
    # @HRModel.remote_callable
    # def models(self, *args, **kwargs):
    #     return self.drsai.a_list_models(*args, **kwargs)

        # request = self.params2request(*args, **kwargs)
        # return self.drsai.a_chat_completions(request=request)


@dataclass
class DrSaiModelConfig(HModelConfig):
    name: str = field(default="hepai/drsai", metadata={"help": "Model's name"})
    permission: Union[str, Dict] = field(default=None, metadata={"help": "Model's permission, separated by ;, e.g., 'groups: all; users: a, b; owner: c', will inherit from worker permissions if not setted"})
    version: str = field(default="2.0", metadata={"help": "Model's version"})

@dataclass
class DrSaiWorkerConfig(HWorkerConfig):
    host: str = field(default="0.0.0.0", metadata={"help": "Worker's address, enable to access from outside if set to `0.0.0.0`, otherwise only localhost can access"})
    port: int = field(default=42801, metadata={"help": "Worker's port, default is None, which means auto start from `auto_start_port`"})
    auto_start_port: int = field(default=42801, metadata={"help": "Worker's start port, only used when port is set to `auto`"})
    route_prefix: str = field(default="/apiv2", metadata={"help": "Route prefix for worker"})
    # controller_address: str = field(default="https://aiapi001.ihep.ac.cn", metadata={"help": "The address of controller"})
    controller_address: str = field(default="http://localhost:42601", metadata={"help": "The address of controller"})
    
    controller_prefix: str = field(default="/apiv2", metadata={"help": "Controller's route prefix"})
    no_register: bool = field(default=True, metadata={"help": "Do not register to controller"})
    

    permissions: str = field(default='groups: default; users: admin, xiongdb@ihep.ac.cn, ddf_free; owner: xiongdb@ihep.ac.cn', metadata={"help": "Model's permissions, separated by ;, e.g., 'groups: default; users: a, b; owner: c'"})
    description: str = field(default='This is Dr.Sai multi agents system', metadata={"help": "Model's description"})
    author: str = field(default=None, metadata={"help": "Model's author"})
    daemon: bool = field(default=False, metadata={"help": "Run as daemon"})
    type: str = field(default="drsai", metadata={"help": "Worker's type"})
    debug: bool = field(default=True, metadata={"help": "Debug mode"})


class Run_DrSaiAPP:
    def __init__(
            self,
            model_args: HModelConfig = DrSaiModelConfig,
            worker_args: HWorkerConfig = DrSaiWorkerConfig,
            **kwargs
            ):
        self.model_args, self.worker_args = hepai.parse_args((model_args, worker_args))

    async def run_drsai(
        self, 
        model_name: str = None,
        host: str = None,
        port: int = None,
        no_register: bool = True,
        controller_address: str = "http://localhost:42601",
        drsaiapp: DrSaiAPP = DrSaiAPP, # 传入DrSaiAPP实例:
        enable_pipeline: bool = False,
        ):
        
        if isinstance(drsaiapp, type): # 传入DrSaiAPP类而不是实例
            drsaiapp = drsaiapp()  # Instantiate the DrSaiAPP instance.

        if model_name is not None:
            self.model_args.name = model_name
        model = DrSaiWorkerModel(config=self.model_args, drsaiapp=drsaiapp)

        if host is not None:
            self.worker_args.host = host
        if port is not None:
            self.worker_args.port = port
        if no_register is not None:
            self.worker_args.no_register = no_register
        self.worker_args.controller_address = controller_address
        
        print(self.model_args)
        print()
        print(self.worker_args)
        print()
        
        if enable_pipeline:
            # 通过Pipeline适配OpenWebUI
            from .owebui_pipeline.api import app as owebui_pipeline_app
            from .owebui_pipeline.api import lifespan as owebui_lifespan
            # 实例化HWorkerAPP
            self.app: FastAPI = HWorkerAPP(
                model, worker_config=self.worker_args,
                lifespan=owebui_lifespan,
                )  # Instantiate the APP, which is a FastAPI application.
            self.app.mount("/pipelines", app=owebui_pipeline_app)
            
        else:
            self.app: FastAPI = HWorkerAPP(
                model, worker_config=self.worker_args
                )
       
        self.app.include_router(model.drsai.router)
      
        print(self.app.worker.get_worker_info(), flush=True)
        # # 启动服务
        # uvicorn.run(self.app, host=self.app.host, port=self.app.port)
        # 创建uvicorn配置和服务实例
        config = uvicorn.Config(
            self.app, 
            host=self.worker_args.host,  # 确保这里使用的是正确的host参数
            port=self.worker_args.port   # 确保这里使用的是正确的port参数
        )
        server = uvicorn.Server(config)
        # 在现有事件循环中启动服务
        if enable_pipeline:
            print(f"Enable OpenWebUI pipelines: `http://{self.worker_args.host}:{self.worker_args.port}/pipelines` with API-KEY: `{owebui_pipeline_app.api_key}`")
        await server.serve()

async def run_console(agent_factory: callable, task: str, **kwargs):
    drsaiapp = DrSaiAPP(agent_factory = agent_factory)
    result = await drsaiapp.start_console(task=task, **kwargs)
    if result is not None:
        return result


async def run_backend(agent_factory: callable, **kwargs):
    '''
    启动后端服务
    : args:
        enable_openwebui_pipeline: bool = False,  # 是否启动openwebui pipelines
    '''
    model_name: str = kwargs.pop("model_name", None)
    host: str =  kwargs.pop("host", None)
    port: int =  kwargs.pop("port", None)
    no_register: bool =  kwargs.pop("no_register", True)
    controller_address: str =  kwargs.pop("controller_address", "http://localhost:42601")
    enable_pipeline: bool = kwargs.pop("enable_openwebui_pipeline", False)
    
    drsaiapp = DrSaiAPP(
        agent_factory = agent_factory,
        **kwargs
        )
    
    await Run_DrSaiAPP().run_drsai(
        model_name=model_name,
        host=host,
        port=port,
        no_register=no_register,
        controller_address=controller_address,
        drsaiapp=drsaiapp,
        enable_pipeline=enable_pipeline,
    )

async def run_hepai_worker(agent_factory: callable, **kwargs):
    '''
    启动Hepai Worker
    '''
    model_name: str = kwargs.pop("model_name", None)
    host: str =  kwargs.pop("host", None)
    port: int =  kwargs.pop("port", None)
    no_register: bool =  kwargs.pop("no_register", False)
    controller_address: str =  kwargs.pop("controller_address", "https://aiapi.ihep.ac.cn")

    drsaiapp = DrSaiAPP(
        agent_factory = agent_factory,
        **kwargs
        )
    
    await Run_DrSaiAPP().run_drsai(
        model_name=model_name,
        host=host,
        port=port,
        no_register=no_register,
        controller_address=controller_address,
        drsaiapp=drsaiapp
    )

async def run_pipelines(pipelines_path: str, pipelines_port: int = 9097):

    NOTES=f"""++++++++++++++++++++++++++++++++\n\n
Open-webui pipelines is starting... \n
If any error, maybe you need to install open-webui first:
```
pip install open-webui
```
Any requriments refrence to https://github.com/open-webui/pipelines"

"""
    print(NOTES)

    # # 获取当前文件夹路径
    # current_path = os.path.dirname(os.path.abspath(__file__))
    # # 获取pipelines文件夹路径./pipelines/main.py
    # pipelines_path = os.path.join(current_path, "pipelines")
    command = f"pm2 start -n pipelines --cwd {pipelines_path} 'python main.py --port {pipelines_port}'"

    print(f"command: {command}")

    # 使用subprocess的asyncio接口来异步执行命令
    process = await asyncio.create_subprocess_shell(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    
    # 获取命令输出
    stdout, stderr = await process.communicate()
    
    # 打印输出信息
    print(f'[Standard output]\n{stdout.decode()}')
    print(f'[Standard error]\n{stderr.decode()}')
    
    # 检查命令是否执行成功
    if process.returncode == 0:
        print(f'Command executed successfully, please visit http://localhost:{pipelines_port}\n"pm2 logs pipelines" to check logs. ')
    else:
        print('Command execution failed')
    
    await asyncio.create_subprocess_shell(
        "pm2 save",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    print("++++++++++++++++++++++++++++++++ \n\n")


async def run_openwebui(openwebui_port: int = 8088):

    NOTES= f'''++++++++++++++++++++++++++++++++\n\n
Open-webui frontend is starting... \n
If any error, maybe you need to install open-webui first:
```
pip install open-webui
```
Any requriments refrence to https://github.com/open-webui/open-webui \n
'''
    print(NOTES)

    # 将huggface镜像站点添加到环境变量：
    os.environ['HF_ENDPOINT'] = 'https://hf-mirror.com'

    # 构建命令
    command = f"pm2 start -n open-webui 'open-webui serve --port {openwebui_port}'"
    
    # 使用subprocess的asyncio接口来异步执行命令
    process = await asyncio.create_subprocess_shell(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    
    # 获取命令输出
    stdout, stderr = await process.communicate()
    
    # 打印输出信息
    print(f'[Standard output]\n{stdout.decode()}')
    print(f'[Standard error]\n{stderr.decode()}')
    
    # 检查命令是否执行成功
    if process.returncode == 0:
        print(f'Command executed successfully, please visit http://localhost:{openwebui_port}\n"pm2 logs open-webui" to check logs. ')
    else:
        print('Command execution failed')
    
    await asyncio.create_subprocess_shell(
        "pm2 save",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    print("++++++++++++++++++++++++++++++++ \n\n")

async def run_drsai_app(agent_factory: callable, pipelines_path: str, **kwargs):
    ''''
    pm2 启动drsai后端/openwebui pipelines/openwebui前端
    '''
    # 检查是否安装了pm2
    try:
        subprocess.check_output("pm2 -v", shell=True)
    except:
        print("Please install pm2 first: `https://pm2.io/`")
        return

    # 并行启动所有服务
    pipelines_port = kwargs.pop("pipelines_port", 9097)
    openwebui_port = kwargs.pop("openwebui_port", 8088)
    await asyncio.gather(
        run_backend(agent_factory=agent_factory, **kwargs),
        run_pipelines(pipelines_path=pipelines_path, pipelines_port=pipelines_port),
        run_openwebui(openwebui_port=openwebui_port)
    )



from dataclasses import dataclass, field
from typing import Optional, List, Dict, Any, Union
import time
from .base_assistant import BaseObject

from hepai.agents.utils import Logger
from .base_run_step import ThreadRunStep

logger = Logger.get_logger("base_run.py")



@dataclass
class Tool:
    type: str

    def to_dict(self):
        return self.__dict__

@dataclass
class Usage:
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int

    def to_dict(self):
        return self.__dict__

@dataclass
class TruncationStrategy:
    type: str
    last_messages: Optional[Any] = None  # 根据实际情况，这里可能需要调整类型

    def to_dict(self):
        return self.__dict__

@dataclass
class ThreadRun(BaseObject):
    id: str
    object: str
    created_at: int
    assistant_id: Union[str, List[str]]
    thread_id: str
    status: str
    started_at: int
    expires_at: Optional[int]
    cancelled_at: Optional[int]
    failed_at: Optional[int]
    completed_at: int
    last_error: Optional[Any]
    model: str
    instructions: Optional[Any]
    incomplete_details: Optional[Any]
    tools: List[Tool]
    metadata: Dict[str, Any]
    usage: Optional[Usage]
    temperature: float
    top_p: float
    max_prompt_tokens: int
    max_completion_tokens: int
    truncation_strategy: TruncationStrategy
    response_format: str
    tool_choice: str

    username: str = field(default=None)

    steps: List = field(default_factory=list)

    def __post_init__(self):
        super(ThreadRun, self).__post_init__()
        self._thread = None
        self._assistants: List = None

    @property
    def thread(self):
        if self._thread is None:
            from hepai.agents import THREADS_MGR
            self._thread = THREADS_MGR.retrieve_thread(self.thread_id, username=self.username)
        return self._thread
    
    @thread.setter
    def thread(self, thread):
        self._thread = thread
    
    @property
    def assistants(self):
        if self._assistants is None:
            from hepai.agents import ASSISTANTS_MGR
            if isinstance(self.assistant_id, str):
                assistant_ids = [self.assistant_id]
            else:
                assistant_ids = self.assistant_id
            # 获取所有的assistants
            self._assistants = [
                ASSISTANTS_MGR.retrieve_assistant(x, username=self.username)
                for x in assistant_ids]
        return self._assistants
    
    @assistants.setter
    def assistants(self, value):
        self._assistants = value
        
    @property
    def output_keys(self):
        """
        不包含thread和assistants字段
        """
        return ["id", "object", "created_at", "assistant_id", "thread_id", "status", "started_at", "expires_at",
                "cancelled_at", "failed_at", "completed_at", "last_error", "model", "instructions", "incomplete_details",
                "tools", "metadata", "usage", "temperature", "top_p", "max_prompt_tokens", "max_completion_tokens",
                "truncation_strategy", "response_format", "tool_choice", "username"]

    def start(self, run_func, **kwargs):
        self.status = "in_progress"
        self.started_at = int(time.time())
        self.thread.start(run_func, run=self, **kwargs)

    def to_dict(self, only_output_keys=True):
        new_dict = dict()
        for k, v in self.__dict__.items():
            if only_output_keys and k not in self.output_keys:
                continue
            if k == "tools" and v:
                v = [x.to_dict() for x in v]
            elif k == "usage" and v:
                v = v.to_dict()
            elif k == "truncation_strategy" and v:
                v = v.to_dict()
            new_dict[k] = v
        return new_dict 
    
    def construct_status_event(self, status=None):
        """根据状态构建事件对象"""
        status = status or self.status
        return {
            "data": self.to_dict(),
            "event": f"thread.run.{status}"
            }

    @property
    def valid_status(self):
        return ["queued", "created", "requires_action", "in_progress", 
                "completed", "failed", "cancelling", "cancelled", "expired"]

    def create_run_step(self, **kwargs):
        run_step_type = kwargs.get("type", "message_creation")
        stream = kwargs.get("stream", False)
        # evc: EventCollector = kwargs.get("event_collector", None)
        assert run_step_type in ["message_creation", "tool_calls"], f"Invalid run step type: {run_step_type}"

        step_id = self.auto_id(prefix="step_", length=30)

        step = ThreadRunStep(
            id=step_id,
            object="thread.run.step",
            created_at=int(time.time()),
            run_id=self.id,
            assistant_id=self.assistant_id,
            thread_id=self.thread_id,
            type=run_step_type,
            status="created",
            step_details=None,
            usage=None
        )

        self.steps.append(step)
        
        # if stream and evc:
        #     evc.add_event_source(step.event_generator())  # 添加事件源
        #     step.set_status("created", emit=True)  # 设置状态，同时触发事件
        return step



"""
基于OAI的SyncAPI的Client
"""

import os
import json
from typing import Mapping, Generator, Dict, AsyncGenerator
from typing_extensions import override
import httpx
import warnings


from .openai_api import *

from .resources._resource import SyncAPIResource
from ._exceptions import HAPIStatusError

# from . import resources
from dataclasses import dataclass, field, asdict

class DemoResoure(SyncAPIResource):
    """
    This is a demo resource for demostrating how to use SyncAPIResource
    
    Usage:
        ```python
        class MyClient(HClient):
            
            def __init__(self):
                super().__init__()
                self.demo_resource: SyncAPIResource = DemoResoure()
        ```

        ```python
        client = MyClient()
        client.demo_resource.get_index()
        ```
    """
    def get_index(self):
        return self._get(
            "/",
            cast_to=Any,
        )
    
    def post_index(self, data):
        return self._post(
            "/",
            cast_to=Any,
            body=data,
        )
    

def get_defualt_timeout(timeout: float = 600.0, connect: float = 5.0) -> httpx.Timeout:
    return httpx.Timeout(timeout=timeout, connect=connect)


# DEFAULT_BASE_URL = os.environ.get("HEPAI_BASE_URL", None)
# DEFAULT_API_KEY = os.environ.get("HEPAI_API_KEY", None)

@dataclass
class HClientConfig:
    # base_url: str = field(default=NOT_GIVEN, metadata={"description": "The default base URL for all requests"})
    base_url: str = field(default="https://aiapi.ihep.ac.cn/apiv2", metadata={"description": "The default base URL for all requests"})
    api_key: str = field(default=NOT_GIVEN, metadata={"description": "The default API key for all requests"})
    max_retries: int = field(default=0, metadata={"description": "The default maximum number of retries for all requests"})
    timeout: None | httpx.Timeout = field(default_factory=get_defualt_timeout, metadata={"description": "The default timeout for all requests"})
    http_client: None | httpx.Client  = field(default=None, metadata={"description": "The default HTTP client for all requests"})
    default_headers: Mapping[str, str] = field(default=None, metadata={"description": "The default headers for all requests"})
    default_query: Mapping[str, object] = field(default=None, metadata={"description": "The default query parameters for all requests"})
    enable_openai: bool = field(default=True, metadata={"description": "Whether to enable openai resources"})
    enable_anthropic: bool = field(default=True, metadata={"description": "Whether to enable anthropic resources"})
    proxy: str = field(default=None, metadata={"description": "The default proxy for all requests"})

    max_connections: int = field(default=1000, metadata={"description": "The maximum number of connections to keep open"})
    max_keepalive_connections: int = field(default=100, metadata={"description": "The maximum number of idle connections to keep open"})
    
    version: str = field(default="2.0.0", metadata={"description": "The version of the client"})
    _strict_response_validation: bool = field(default=False, metadata={"description": "Whether to strictly validate responses"})
    
    def to_dict(self):
        return asdict(self)
    
    def update_by_dict(self, d: dict):
        unknown_keys = []
        for k, v in d.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                unknown_keys.append(k)
        if unknown_keys:
            warnings.warn(f"[HClientConfig] encountered {len(unknown_keys) } unknown keys when updating config, this keys will be ignored: {unknown_keys}")
        return self
    
    def instantiate_http_client(self,) -> httpx.Client:
        """如果http_client为None，则实例化一个httpx.Client"""
        if self.http_client is None:
            # proxies = None if self.proxy is None else {"https://": self.proxy, "http://": self.proxy}
            limits = httpx.Limits(max_connections=self.max_connections, max_keepalive_connections=self.max_keepalive_connections)
            self.http_client = httpx.Client(
                base_url=self.base_url,
                timeout=self.timeout,
                # proxies=proxies,
                proxy=self.proxy,
                transport=None,
                limits=limits,
                follow_redirects=True,
            )
        return self.http_client


class HClient(SyncAPIClient):
    """
    高能AI框架基础客户端
    """
    NotGiven = NOT_GIVEN

    def __init__(
            self,
            config: HClientConfig = None,
            **overrides, # 用于覆盖默认配置
    ):
        self.config = config or HClientConfig()
        if overrides:
            self.config.update_by_dict(overrides)
        
        api_key = self.config.api_key
        if api_key == NOT_GIVEN:  # 人为设置为NOT_GIVEN时，不报错
            pass
        elif api_key is None:
            raise ValueError(
                "The api_key client option must be set either by passing api_key to the client or by setting the environment variable"
            )
        self.api_key = api_key


        if self.config.base_url == NOT_GIVEN:
            self.config.base_url = httpx.URL("")
            pass
        elif self.config.base_url is None:
            raise ValueError(
                "The base_url client option must be set, you can set it by passing base_url to the client or by setting the environment variable"
            )
        
        if (self.config.proxy is not None) and (self.config.http_client is None):
            print(f'[HClient] Proxy: {self.config.proxy}')
            self.config.instantiate_http_client()  # 提前实例化http_client，就设置了proxy

        super().__init__(
            version=self.config.version,
            base_url=self.config.base_url,
            max_retries=self.config.max_retries,
            timeout=self.config.timeout,
            http_client=self.config.http_client,
            custom_headers=self.config.default_headers,
            custom_query=self.config.default_query,
            _strict_response_validation=self.config._strict_response_validation,
        )

        if self.config.enable_openai:
            """集成openai的resources"""
            # from openai import OpenAI
            # from openai import resources
            # from openai._client import OpenAIWithRawResponse, OpenAIWithStreamedResponse
            from .openai_api import resources
            # self._default_stream_cls = HClient.Stream
            self.completions = resources.Completions(self)
            self.chat = resources.Chat(self)
            self.embeddings = resources.Embeddings(self)
            self.files = resources.Files(self)
            self.images = resources.Images(self)
            self.audio = resources.Audio(self)
            self.moderations = resources.Moderations(self)
            self.models = resources.Models(self)
            self.fine_tuning = resources.FineTuning(self)
            self.beta = resources.Beta(self)
            self.batches = resources.Batches(self)
            self.uploads = resources.Uploads(self)
            # self.with_raw_response = OpenAIWithRawResponse(self)
            # self.with_streaming_response = OpenAIWithStreamedResponse(self)

        if self.config.enable_anthropic:
            """集成anthropic的resources"""
            # from openai import resources
            # from openai._client import OpenAIWithRawResponse, OpenAIWithStreamedResponse
            # from .anthropic_api import resources
            # # self._default_stream_cls = HClient.Stream
            # self.anthropic = resources.Anthropic(self)
            # self.with_raw_response = OpenAIWithRawResponse(self)
            # self.with_streaming_response = OpenAIWithStreamedResponse(self)
            # from anthropic import Anthropic
            try:
                import anthropic
            except ImportError:
                raise ImportError(
                    "Please install the `anthropic` package to use the Anthropic resources, you can install it by `pip install anthropic`"
                )
            self._anthropic: anthropic.Anthropic | None = None
            self._anthropic_params = {
                "api_key": self.api_key,
                "base_url": self.config.base_url,  # type: ignore[call-arg]
                "timeout": self.config.timeout,  # type: ignore[call-arg]
                "max_retries": self.config.max_retries,  # type: ignore[call-arg]
                "http_client": self.config.http_client,  # type: ignore[call-arg]
                "default_headers": self.config.default_headers,
                "default_query": self.config.default_query,
                "_strict_response_validation": self.config._strict_response_validation,  # type: ignore[call-arg]
            }
                
    
    @property
    def anthropic(self):
        if self._anthropic is None:
            from anthropic import Anthropic
            self._anthropic = Anthropic(**self._anthropic_params)
        return self._anthropic
    
    @property
    def Stream(self):
        return Stream
    
    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="brackets")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        if api_key is None:
            return {}
        return {"Authorization": f"Bearer {api_key}"}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": "false",
            **self._custom_headers,
        }
    
    def stream_to_generator(self, stream_obj: "Stream") -> Generator:
        """make a stream object to a generator that fit to client stream decoder"""
        for x in stream_obj:
            # print(x)
            yield f"data: {json.dumps(x)}\n\n"

    def _make_status_error(self, err_msg: str, *, body: object, response: httpx.Response) -> HAPIStatusError:
        """
        Make an APIStatusError from an error message, response body, and response object.
        For example: 
            err_msg: str, "Error code: 401 - {'detail': 'API-KEY not provied, please provide API Key in the header by set `Authorization` header'}"
            body: dict, {'detail': 'API-KEY not provied, please provide API Key in the header by set `Authorization` header'}
            response: httpx.Response, <Response [401]>
        """
        return HAPIStatusError(err_msg, body=body, response=response)
        
        
        # return super()._make_status_error(err_msg, body=body, response=response)


class AsyncHClient(AsyncAPIClient):
    """
    高能AI框架基础异步客户端
    """
    NotGiven = NOT_GIVEN

    def __init__(
            self,
            config: HClientConfig = None,
            **overrides,  # 用于覆盖默认配置
    ):
        self.config = config or HClientConfig()
        if overrides:
            self.config.update_by_dict(overrides)
        
        api_key = self.config.api_key
        if api_key == NOT_GIVEN:  # 人为设置为NOT_GIVEN时，不报错
            pass
        elif api_key is None:
            raise ValueError(
                "The api_key client option must be set either by passing api_key to the client or by setting the environment variable"
            )
        self.api_key = api_key

        if self.config.base_url == NOT_GIVEN:
            self.config.base_url = httpx.URL("")
            pass
        elif self.config.base_url is None:
            raise ValueError(
                "The base_url client option must be set, you can set it by passing base_url to the client or by setting the environment variable"
            )
        
        if (self.config.proxy is not None) and (self.config.http_client is None):
            print(f'[AsyncHClient] Proxy: {self.config.proxy}')
            self.config.instantiate_http_client()  # 提前实例化http_client，就设置了proxy

        super().__init__(
            version=self.config.version,
            base_url=self.config.base_url,
            max_retries=self.config.max_retries,
            timeout=self.config.timeout,
            http_client=self.config.http_client,
            custom_headers=self.config.default_headers,
            custom_query=self.config.default_query,
            _strict_response_validation=self.config._strict_response_validation,
        )

        if self.config.enable_openai:
            """集成openai的resources"""
            from .openai_api import resources
            self.completions = resources.AsyncCompletions(self)
            self.chat = resources.AsyncChat(self)
            self.embeddings = resources.AsyncEmbeddings(self)
            self.files = resources.AsyncFiles(self)
            self.images = resources.AsyncImages(self)
            self.audio = resources.AsyncAudio(self)
            self.moderations = resources.AsyncModerations(self)
            self.models = resources.AsyncModels(self)
            self.fine_tuning = resources.AsyncFineTuning(self)
            self.beta = resources.AsyncBeta(self)
            self.batches = resources.AsyncBatches(self)
            self.uploads = resources.AsyncUploads(self)

    @property
    def Stream(self):
        # return Stream
        return AsyncStream
    
    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="brackets")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        if api_key is None:
            return {}
        return {"Authorization": f"Bearer {api_key}"}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": "false",
            **self._custom_headers,
        }

    async def stream_to_generator(self, stream_obj: "Stream") -> AsyncGenerator:
        """Make a stream object to an async generator that fits the client stream decoder"""
        async for x in stream_obj:
            yield f"data: {json.dumps(x)}\n\n"

    def _make_status_error(self, err_msg: str, *, body: object, response: httpx.Response) -> HAPIStatusError:
        """
        Make an APIStatusError from an error message, response body, and response object.
        """
        return HAPIStatusError(err_msg, body=body, response=response)



if __name__ == "__main__":
    client = HClient(base_url="http://localhost:42600/apiv2")    
    
    from hepai.components.haiddf.hclient.openai_api import Stream
    from typing import Any
    data_need_stream = [
            1, 2, 3,
            "x", "y", "z",
            [[1, 2], [3, 4], [5, 6]],
            {"a": "b", "c": "d"},
        ]
    
    rst = client.post(
            path="/worker_unified_gate/?function=get_stream", 
            cast_to=Any,
            body={"kwargs": {"data": data_need_stream, "interval": 0.1}},
            stream=True,
            stream_cls=Stream[Any],
            )
    for i, x in enumerate(rst):
        print(f"i: {i}, x: {x}, type: {type(x)}")